from unittest import mock
import unittest
import sys
import os
import shutil
import io
import zipfile
from flask import current_app
import webscrapbook
from webscrapbook import WSB_DIR, WSB_LOCAL_CONFIG
from webscrapbook import app as wsbapp

root_dir = os.path.abspath(os.path.dirname(__file__))
server_root = os.path.join(root_dir, 'test_app_helpers')

mocking = None

def setUpModule():
    # mock out WSB_USER_CONFIG
    global mocking
    mocking = mock.patch('webscrapbook.WSB_USER_CONFIG', server_root)
    mocking.start()

def tearDownModule():
    # stop mock
    mocking.stop()

class TestFunctions(unittest.TestCase):
    def test_is_local_access(self):
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        app = wsbapp.make_app(root)

        # host is localhost
        with app.test_request_context('/',
                base_url='http://127.0.0.1',
                environ_base={'REMOTE_ADDR': '192.168.0.100'}):
            self.assertTrue(wsbapp.is_local_access())

        # host (with port) is localhost
        with app.test_request_context('/',
                base_url='http://127.0.0.1:8000',
                environ_base={'REMOTE_ADDR': '192.168.0.100'}):
            self.assertTrue(wsbapp.is_local_access())

        # remote is localhost
        with app.test_request_context('/',
                base_url='http://192.168.0.1',
                environ_base={'REMOTE_ADDR': '127.0.0.1'}):
            self.assertTrue(wsbapp.is_local_access())

        # host = remote
        with app.test_request_context('/',
                base_url='http://example.com',
                environ_base={'REMOTE_ADDR': 'example.com'}):
            self.assertTrue(wsbapp.is_local_access())

        # host (with port) = remote
        with app.test_request_context('/',
                base_url='http://example.com:8000',
                environ_base={'REMOTE_ADDR': 'example.com'}):
            self.assertTrue(wsbapp.is_local_access())

        # otherwise non-local
        with app.test_request_context('/',
                base_url='http://example.com',
                environ_base={'REMOTE_ADDR': '192.168.0.100'}):
            self.assertFalse(wsbapp.is_local_access())

    def test_get_archive_path1(self):
        """Basic logit for a sub-archive path."""
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        app = wsbapp.make_app(root)
        with app.app_context():
            tempfile = os.path.join(root, 'entry.zip')
            try:
                with zipfile.ZipFile(tempfile, 'w') as zip:
                    pass

                self.assertEqual(wsbapp.get_archive_path('entry.zip'), (None, None))
                self.assertEqual(wsbapp.get_archive_path('entry.zip!'), (None, None))
                self.assertEqual(wsbapp.get_archive_path('entry.zip!/'), (tempfile, ''))
                self.assertEqual(wsbapp.get_archive_path('entry.zip!/subdir'), (tempfile, 'subdir'))
                self.assertEqual(wsbapp.get_archive_path('entry.zip!/subdir/'), (tempfile, 'subdir'))
                self.assertEqual(wsbapp.get_archive_path('entry.zip!/index.html'), (tempfile, 'index.html'))
            finally:
                try:
                    os.remove(tempfile)
                except FileNotFoundError:
                    pass

    def test_get_archive_path2(self):
        """Handle conflicting file or directory."""
        # entry.zip!/entry1.zip!/ = entry.zip!/entry1.zip! >
        # entry.zip!/entry1.zip >
        # entry.zip!/ = entry.zip! >
        # entry.zip
        root = os.path.join(root_dir, 'test_app_helpers', 'general')
        app = wsbapp.make_app(root)
        with app.app_context():
            # entry.zip!/entry1.zip!/ > entry.zip!/entry1.zip
            try:
                os.makedirs(os.path.join(root, 'entry.zip!', 'entry1.zip!'), exist_ok=True)
                with zipfile.ZipFile(os.path.join(root, 'entry.zip!', 'entry1.zip'), 'w') as zip:
                    pass
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('entry.zip!/entry1.zip!/'),
                    (None, None))
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # entry.zip!/entry1.zip! > entry.zip!/entry1.zip
            try:
                os.makedirs(os.path.join(root, 'entry.zip!'), exist_ok=True)
                with open(os.path.join(root, 'entry.zip!', 'entry1.zip!'), 'w') as f:
                    pass
                with zipfile.ZipFile(os.path.join(root, 'entry.zip!', 'entry1.zip'), 'w') as zip:
                    pass
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('entry.zip!/entry1.zip!/'),
                    (None, None))
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # entry.zip!/entry1.zip > entry.zip!/
            try:
                os.makedirs(os.path.join(root, 'entry.zip!'), exist_ok=True)
                with zipfile.ZipFile(os.path.join(root, 'entry.zip!', 'entry1.zip'), 'w') as zip:
                    pass
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('entry.zip!/entry1.zip!/'),
                    (os.path.join(root, 'entry.zip!', 'entry1.zip'), ''))
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # entry.zip!/ > entry.zip
            try:
                os.makedirs(os.path.join(root, 'entry.zip!'), exist_ok=True)
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('entry.zip!/entry1.zip!/'),
                    (None, None))
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # entry.zip! > entry.zip
            try:
                with open(os.path.join(root, 'entry.zip!'), 'w') as f:
                    pass
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('entry.zip!/entry1.zip!/'),
                    (None, None))
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # entry.zip
            try:
                with zipfile.ZipFile(os.path.join(root, 'entry.zip'), 'w') as zip:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('entry.zip!/entry1.zip!/'),
                    (os.path.join(root, 'entry.zip'), 'entry1.zip!'))
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

            # other
            try:
                with open(os.path.join(root, 'entry.zip'), 'w') as f:
                    pass

                self.assertEqual(
                    wsbapp.get_archive_path('entry.zip!/entry1.zip!/'),
                    (None, None))
            finally:
                try:
                    shutil.rmtree(os.path.join(root, 'entry.zip!'))
                except NotADirectoryError:
                    os.remove(os.path.join(root, 'entry.zip!'))
                except FileNotFoundError:
                    pass
                try:
                    os.remove(os.path.join(root, 'entry.zip'))
                except FileNotFoundError:
                    pass

    @mock.patch('webscrapbook.util.encrypt', side_effect=webscrapbook.util.encrypt)
    def test_get_permission1(self, mock_encrypt):
        """Return 'all' anyway if no auth section."""
        root = os.path.join(root_dir, 'test_app_helpers', 'get_permission1')
        app = wsbapp.make_app(root)
        with app.app_context():
            # util.encrypt should NOT be called
            self.assertEqual(wsbapp.get_permission(None), 'all')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': '', 'password': ''}), 'all')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': '', 'password': 'pass'}), 'all')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': 'user', 'password': ''}), 'all')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': 'user', 'password': 'pass'}), 'all')
            mock_encrypt.assert_not_called()

    @mock.patch('webscrapbook.util.encrypt', side_effect=webscrapbook.util.encrypt)
    def test_get_permission2(self, mock_encrypt):
        """Return corresponding permission for the matched user and '' for unmatched."""
        root = os.path.join(root_dir, 'test_app_helpers', 'get_permission2')
        app = wsbapp.make_app(root)
        with app.app_context():
            # util.encrypt should be called with the inputting password
            # and the salt and method for the matched user
            mock_encrypt.reset_mock()

            self.assertEqual(wsbapp.get_permission({'username': 'user1', 'password': 'pass1'}), '')
            mock_encrypt.assert_called_with('pass1', '', 'plain')

            self.assertEqual(wsbapp.get_permission({'username': 'user2', 'password': 'pass2'}), 'view')
            mock_encrypt.assert_called_with('pass2', 'salt', 'plain')

            self.assertEqual(wsbapp.get_permission({'username': 'user3', 'password': 'pass3'}), 'read')
            mock_encrypt.assert_called_with('pass3', '', 'sha1')

            self.assertEqual(wsbapp.get_permission({'username': 'user4', 'password': 'pass4'}), 'all')
            mock_encrypt.assert_called_with('pass4', 'salt4', 'sha256')

            # Password check should be handled by util.encrypt properly.
            # Here are just some quick fail tests for certain cases:
            # - empty input should not work
            # - inputting password + salt should not work
            # - inputting hashed value should not work
            mock_encrypt.reset_mock()

            self.assertEqual(wsbapp.get_permission({'username': 'user4', 'password': ''}), '')
            mock_encrypt.assert_called_with('', 'salt4', 'sha256')

            self.assertEqual(wsbapp.get_permission({'username': 'user4', 'password': 'salt4'}), '')
            mock_encrypt.assert_called_with('salt4', 'salt4', 'sha256')

            self.assertEqual(wsbapp.get_permission({'username': 'user4', 'password': '49d1445a2989c509c5b5b1f78e092e3f30f05b1d219fd975ac77ff645ea68d53'}), '')
            mock_encrypt.assert_called_with('49d1445a2989c509c5b5b1f78e092e3f30f05b1d219fd975ac77ff645ea68d53', 'salt4', 'sha256')

            # util.encrypt should NOT be called for an unmatched user
            mock_encrypt.reset_mock()

            self.assertEqual(wsbapp.get_permission(None), '')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': '', 'password': ''}), '')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': '', 'password': 'pass'}), '')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': 'userx', 'password': ''}), '')
            mock_encrypt.assert_not_called()

            self.assertEqual(wsbapp.get_permission({'username': 'userx', 'password': 'pass'}), '')
            mock_encrypt.assert_not_called()

    @mock.patch('webscrapbook.util.encrypt', side_effect=webscrapbook.util.encrypt)
    def test_get_permission3(self, mock_encrypt):
        """Use empty user and password if not provided."""
        root = os.path.join(root_dir, 'test_app_helpers', 'get_permission3')
        app = wsbapp.make_app(root)
        with app.app_context():
            self.assertEqual(wsbapp.get_permission(None), 'view')
            mock_encrypt.assert_called_with('', 'salt', 'plain')

            self.assertEqual(wsbapp.get_permission({'username': '', 'password': ''}), 'view')
            mock_encrypt.assert_called_with('', 'salt', 'plain')

    @mock.patch('webscrapbook.util.encrypt', side_effect=webscrapbook.util.encrypt)
    def test_get_permission4(self, mock_encrypt):
        """Use permission for the first matched user and password."""
        root = os.path.join(root_dir, 'test_app_helpers', 'get_permission4')
        app = wsbapp.make_app(root)
        with app.app_context():
            mock_encrypt.reset_mock()
            self.assertEqual(wsbapp.get_permission({'username': '', 'password': ''}), 'view')
            mock_encrypt.assert_called_once_with('', 'salt', 'plain')

            mock_encrypt.reset_mock()
            self.assertEqual(wsbapp.get_permission({'username': 'user1', 'password': 'pass1'}), 'read')
            self.assertEqual(mock_encrypt.call_args_list[0][0], ('pass1', 'salt', 'plain'))
            self.assertEqual(mock_encrypt.call_args_list[1][0], ('pass1', 'salt', 'plain'))

    def test_verify_authorization(self):
        for action in {'view', 'source', 'static'}:
            with self.subTest(action=action):
                self.assertFalse(wsbapp.verify_authorization('', action))
                self.assertTrue(wsbapp.verify_authorization('view', action))
                self.assertTrue(wsbapp.verify_authorization('read', action))
                self.assertTrue(wsbapp.verify_authorization('all', action))

        for action in {'list', 'edit', 'editx', 'exec', 'browse', 'config', 'unknown'}:
            with self.subTest(action=action):
                self.assertFalse(wsbapp.verify_authorization('', action))
                self.assertFalse(wsbapp.verify_authorization('view', action))
                self.assertTrue(wsbapp.verify_authorization('read', action))
                self.assertTrue(wsbapp.verify_authorization('all', action))

        for action in {'token', 'lock', 'unlock', 'mkdir', 'save', 'delete', 'move', 'copy'}:
            with self.subTest(action=action):
                self.assertFalse(wsbapp.verify_authorization('', action))
                self.assertFalse(wsbapp.verify_authorization('view', action))
                self.assertFalse(wsbapp.verify_authorization('read', action))
                self.assertTrue(wsbapp.verify_authorization('all', action))

    def test_make_app1(self):
        # pass root
        root = os.path.join(root_dir, 'test_app_helpers', 'make_app1')

        app = wsbapp.make_app(root)
        with app.app_context():
            self.assertEqual(current_app.config['WEBSCRAPBOOK_RUNTIME']['config']['app']['name'], 'mywsb1')

    def test_make_app2(self):
        # pass root, config
        root = os.path.join(root_dir, 'test_app_helpers', 'make_app1')
        config_dir = os.path.join(root_dir, 'test_app_helpers', 'make_app2')
        config = webscrapbook.Config()
        config.load(config_dir)

        app = wsbapp.make_app(root, config)
        with app.app_context():
            self.assertEqual(current_app.config['WEBSCRAPBOOK_RUNTIME']['config']['app']['name'], 'mywsb2')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# time.time_ns is available since Python 3.7
try:
    from time import time_ns
except ImportError:
    from time import time
    def time_ns():
        return int(time() * 1e9)



#!/usr/bin/env python3
"""A backend toolkit for management of WebScrapBook collection.
"""
import sys
import os
from configparser import ConfigParser
from collections import OrderedDict
from copy import deepcopy
import mimetypes
import re

__all__ = ['WSB_USER_CONFIG', 'WSB_DIR', 'WSB_LOCAL_CONFIG', 'config']

__package_name__ = 'webscrapbook'
__version__ = '0.17.1'
__author__ = 'Danny Lin'
__author_email__ = 'danny0838@gmail.com'
__homepage__ = 'https://github.com/danny0838/PyWebScrapBook'
__license__ = 'MIT'

WSB_USER_CONFIG = os.environ.get('WSB_USER_CONFIG') or os.path.join(os.path.expanduser('~'), '.wsbconfig')
WSB_DIR = os.environ.get('WSB_DIR') or '.wsb'
WSB_LOCAL_CONFIG = os.environ.get('WSB_LOCAL_CONFIG') or 'config.ini'

mimetypes.add_type("application/rss+xml", ".rss")
mimetypes.add_type("application/atom+xml", ".atom")
mimetypes.add_type("text/markdown", ".md")
mimetypes.add_type("text/markdown", ".mkd")
mimetypes.add_type("text/markdown", ".mkdn")
mimetypes.add_type("text/markdown", ".mdwn")
mimetypes.add_type("text/markdown", ".mdown")
mimetypes.add_type("text/markdown", ".markdown")
mimetypes.add_type("application/html+zip", ".htz")
mimetypes.add_type("application/x-maff", ".maff")


class Config():
    """Config class whose values are lazily-initialized when accessed.

    Values are loaded from config files relative to CWD. Consider calling
    load() if the script changes CWD during the runtime.
    """
    def __init__(self):
        self._conf = None
        self._data = None


    def __getitem__(self, key):
        if self._conf is None: self.load()  # lazy load
        return self._data[key]


    def __iter__(self):
        if self._conf is None: self.load()  # lazy load
        return iter(self._data)


    def getname(self, name):
        if self._conf is None: self.load()  # lazy load
        parts = name.split('.')
        if len(parts) == 3:
            sec, subsec, key = parts
            try:
                return self._conf['{} "{}"'.format(sec, subsec)][key]
            except KeyError:
                pass
        elif len(parts) == 2:
            sec, key = parts
            try:
                return self._conf[sec][key]
            except KeyError:
                pass
        return None


    def dump(self, fh):
        if self._conf is None: self.load()  # lazy load
        self._conf.write(fh)


    def dump_object(self):
        """Dump configs as an object, with type casting.
        """
        if self._conf is None: self.load()  # lazy load
        return deepcopy(self._data)


    def load(self, root='.'):
        """Loads config files related to the given root directory.

        Skip if config file doesn't exist, but raise if not loadable.

        project config > user config > default config
        """
        def load_config(file):
            if not os.path.isfile(file):
                return

            try:
                parser = ConfigParser(interpolation=None)
                parser.read(file, encoding='UTF-8')
                for section in parser.sections():
                    # Handle subsected sections formatted as [section "subsection"].
                    # Also normalize [section] and [section  ""  ] to [section ""].
                    m = re.search(r'^(\S*)(?:\s*"([^"\]]*)"\s*)?$', section)
                    if m:
                        sec, subsec = m.group(1), m.group(2) or ''
                        if sec in self.SUBSECTED:
                            newsection = '{} "{}"'.format(sec, subsec)
                            if newsection != section:
                                conf.setdefault(newsection, OrderedDict())
                                conf[newsection].update(parser[section])
                                continue

                    # conf.setdefault(...).update(...) doesn't work here as the
                    # setdefault may return the default value rather then a
                    # Section object.
                    conf.setdefault(section, OrderedDict())
                    conf[section].update(parser[section])
            except:
                print('Error: Unable to load config from "{}".'.format(file), file=sys.stderr)
                raise

        # default config
        self._conf = conf = ConfigParser(interpolation=None)
        conf['app'] = {}
        conf['app']['name'] = 'WebScrapBook'
        conf['app']['theme'] = 'default'
        conf['app']['root'] = '.'
        conf['app']['base'] = ''
        conf['app']['allowed_x_for'] = '0'
        conf['app']['allowed_x_proto'] = '0'
        conf['app']['allowed_x_host'] = '0'
        conf['app']['allowed_x_port'] = '0'
        conf['app']['allowed_x_prefix'] = '0'
        conf['server'] = {}
        conf['server']['port'] = '8080'
        conf['server']['host'] = 'localhost'
        conf['server']['ssl_on'] = 'false'
        conf['server']['ssl_key'] = ''
        conf['server']['ssl_cert'] = ''
        conf['server']['ssl_pw'] = ''
        conf['server']['browse'] = 'true'
        conf['browser'] = {}
        conf['browser']['command'] = ''
        conf['browser']['index'] = ''
        conf['browser']['cache_prefix'] = 'webscrapbook.'
        conf['browser']['cache_expire'] = '259200'
        conf['browser']['use_jar'] = 'false'
        conf['book ""'] = {}
        conf['book ""']['name'] = 'scrapbook'
        conf['book ""']['top_dir'] = ''
        conf['book ""']['data_dir'] = ''
        conf['book ""']['tree_dir'] = '.wsb/tree'
        conf['book ""']['index'] = '.wsb/tree/map.html'
        conf['book ""']['no_tree'] = 'false'

        # user config
        load_config(WSB_USER_CONFIG)

        # book config
        load_config(os.path.join(root, WSB_DIR, WSB_LOCAL_CONFIG))

        # map subsections
        self._data = OrderedDict()
        for section in conf.sections():
            sectionobj = OrderedDict()
            m = re.search(r'^(\S*)(?:\s*"([^"\]]*)"\s*)?$', section)
            if m:
                sec, subsec = m.group(1), m.group(2) or ''
                if sec in self.SUBSECTED:
                    self._data.setdefault(sec, OrderedDict())[subsec] = sectionobj
                    for key in conf[section]:
                        try:
                            sectionobj[key] = getattr(conf[section], self.TYPES[sec][None][key])(key)
                        except KeyError:
                            sectionobj[key] = conf[section][key]
                    continue
            self._data[section] = sectionobj
            for key in conf[section]:
                try:
                    sectionobj[key] = getattr(conf[section], self.TYPES[section][key])(key)
                except KeyError:
                    sectionobj[key] = conf[section][key]

    SUBSECTED = ['book', 'auth']
    TYPES = {
        'app': {
            'allowed_x_for': 'getint',
            'allowed_x_proto': 'getint',
            'allowed_x_host': 'getint',
            'allowed_x_port': 'getint',
            'allowed_x_prefix': 'getint',
            },
        'server': {
            'port': 'getint',
            'ssl_on': 'getboolean',
            'browse': 'getboolean',
            },
        'browser': {
            'cache_expire': 'getint',
            'use_jar': 'getboolean',
            },
        'book': {
            None: {
                'no_tree': 'getboolean',
                },
            },
        }

config = Config()



from unittest import mock
import unittest
import os
import io
import webscrapbook

root_dir = os.path.abspath(os.path.dirname(__file__))

class TestClassConfig(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self._WSB_USER_CONFIG = webscrapbook.WSB_USER_CONFIG

        # set WSB_USER_CONFIG to a known folder to ensure it's not loaded
        webscrapbook.WSB_USER_CONFIG = os.path.join(root_dir, 'test_config', '.wsb')

    @classmethod
    def tearDownClass(self):
        # recover constants
        webscrapbook.WSB_USER_CONFIG = self._WSB_USER_CONFIG

    def test_load(self):
        conf = webscrapbook.Config()
        conf.load(os.path.join(root_dir, 'test_config'))
        self.assertEqual(conf['app']['name'], 'mywsb')
        self.assertEqual(conf['app']['theme'], 'mytheme')
        self.assertEqual(conf['app']['root'], 'myroot')
        self.assertEqual(conf['app']['base'], 'mybase')
        self.assertEqual(conf['app']['allowed_x_for'], 1)
        self.assertEqual(conf['app']['allowed_x_proto'], 1)
        self.assertEqual(conf['app']['allowed_x_host'], 0)
        self.assertEqual(conf['app']['allowed_x_port'], 0)
        self.assertEqual(conf['app']['allowed_x_prefix'], 0)
        self.assertEqual(conf['server']['ssl_on'], True)
        self.assertEqual(conf['server']['browse'], True)
        self.assertEqual(conf['browser']['use_jar'], False)
        self.assertEqual(conf['book']['']['name'], 'mybook')
        self.assertEqual(conf['book']['']['no_tree'], False)
        self.assertEqual(conf['book']['book2']['name'], 'mybook2')
        self.assertEqual(conf['book']['book2']['no_tree'], True)
        self.assertEqual(conf['auth']['user1']['user'], 'myuser1')

    def test_load_repeated(self):
        conf = webscrapbook.Config()
        conf.load(os.path.join(root_dir, 'test_config'))

        # check if previous loaded config entries no more exist
        conf.load(os.path.join(root_dir, 'test_config_load_repeated'))
        self.assertEqual(conf['app']['name'], 'mywsb2')
        with self.assertRaises(KeyError):
            self.assertEqual(conf['book']['book2']['name'], 'mybook2')
        with self.assertRaises(KeyError):
            self.assertEqual(conf['auth']['user1']['user'], 'myuser1')

    @mock.patch('webscrapbook.WSB_LOCAL_CONFIG', 'localconfig.ini')
    @mock.patch('webscrapbook.WSB_DIR', '.wsbdir')
    @mock.patch('webscrapbook.WSB_USER_CONFIG', os.path.join(root_dir, 'test_config_load_constants', 'userconfig.ini'))
    def test_load_constants(self):
        # check if WSB_USER_CONFIG, WSB_DIR, and WSB_LOCAL_CONFIG are honored
        conf = webscrapbook.Config()
        conf.load(os.path.join(root_dir, 'test_config_load_constants'))
        self.assertEqual(conf['app']['name'], 'myuserwsb')
        self.assertEqual(conf['app']['theme'], 'mytheme')
        self.assertEqual(conf['server']['port'], 8888)
        self.assertEqual(conf['book']['']['name'], 'myuserbook')
        self.assertEqual(conf['book']['']['no_tree'], True)
        self.assertEqual(conf['book']['book1']['name'], 'mybook1')
        self.assertEqual(conf['book']['book2']['name'], 'mybook2')

    def test_getitem(self):
        # test lazy loading
        _cwd = os.getcwd()
        os.chdir(os.path.join(root_dir, 'test_config'))

        try:
            conf = webscrapbook.Config()
            self.assertEqual(conf['app']['name'], 'mywsb')
            self.assertEqual(conf['app']['allowed_x_for'], 1)
            self.assertEqual(conf['server']['ssl_on'], True)
            self.assertEqual(conf['server']['browse'], True)
            self.assertEqual(conf['book']['']['name'], 'mybook')
            self.assertEqual(conf['book']['book2']['name'], 'mybook2')
        finally:
            os.chdir(_cwd)

    def test_iter(self):
        # test lazy loading
        _cwd = os.getcwd()
        os.chdir(os.path.join(root_dir, 'test_config'))

        try:
            conf = webscrapbook.Config()
            self.assertEqual(list(iter(conf)), ['app', 'server', 'browser', 'book', 'auth'])
        finally:
            os.chdir(_cwd)

    def test_getname(self):
        # test lazy loading
        _cwd = os.getcwd()
        os.chdir(os.path.join(root_dir, 'test_config'))

        try:
            conf = webscrapbook.Config()
            self.assertEqual(conf.getname('app.name'), 'mywsb')
            self.assertEqual(conf.getname('app.allowed_x_for'), '1')
            self.assertEqual(conf.getname('server.ssl_on'), 'true')
            self.assertEqual(conf.getname('server.browse'), 'yes')
            self.assertEqual(conf.getname('book..name'), 'mybook')
            self.assertEqual(conf.getname('book.book2.name'), 'mybook2')
        finally:
            os.chdir(_cwd)

    def test_dump(self):
        # test lazy loading
        _cwd = os.getcwd()
        os.chdir(os.path.join(root_dir, 'test_config'))

        try:
            conf = webscrapbook.Config()
            with io.StringIO() as fh:
                conf.dump(fh)
                output = fh.getvalue()
            self.assertEqual(output, """[app]
name = mywsb
theme = mytheme
root = myroot
base = mybase
allowed_x_for = 1
allowed_x_proto = 1
allowed_x_host = 0
allowed_x_port = 0
allowed_x_prefix = 0

[server]
port = 9999
host = localhost
ssl_on = true
ssl_key = ./wsb/wsb.key
ssl_cert = ./wsb/wsb.crt
ssl_pw = 
browse = yes

[browser]
command = 
index = 
cache_prefix = wsb.
cache_expire = 123456
use_jar = no

[book ""]
name = mybook
top_dir = 
data_dir = 
tree_dir = .wsb/tree
index = .wsb/tree/map.html
no_tree = false

[book "book2"]
name = mybook2
no_tree = on

[auth "user1"]
user = myuser1
permission = all

""")
        finally:
            os.chdir(_cwd)

    def test_dump_object(self):
        # test lazy loading
        _cwd = os.getcwd()
        os.chdir(os.path.join(root_dir, 'test_config'))

        try:
            conf = webscrapbook.Config()
            dump = conf.dump_object()
            self.assertEqual(dump['app']['name'], 'mywsb')
            self.assertEqual(dump['app']['allowed_x_for'], 1)
            self.assertEqual(dump['app']['allowed_x_proto'], 1)
            self.assertEqual(dump['app']['allowed_x_host'], 0)
            self.assertEqual(dump['app']['allowed_x_port'], 0)
            self.assertEqual(dump['app']['allowed_x_prefix'], 0)
            self.assertEqual(dump['server']['port'], 9999)
            self.assertEqual(dump['server']['ssl_on'], True)
            self.assertEqual(dump['server']['browse'], True)
            self.assertEqual(dump['browser']['cache_expire'], 123456)
            self.assertEqual(dump['browser']['use_jar'], False)
            self.assertEqual(dump['book']['']['name'], 'mybook')
            self.assertEqual(dump['book']['']['no_tree'], False)
            self.assertEqual(dump['book']['book2']['name'], 'mybook2')
            self.assertEqual(dump['book']['book2']['no_tree'], True)
            self.assertEqual(dump['auth']['user1']['user'], 'myuser1')
        finally:
            os.chdir(_cwd)

if __name__ == '__main__':
    unittest.main()



from unittest import mock
import unittest
import sys
import os
import shutil
import io
import zipfile
import re
from urllib.request import pathname2url
from tempfile import gettempdir
from webscrapbook import WSB_DIR, WSB_LOCAL_CONFIG
from webscrapbook import Config
from webscrapbook import cli

root_dir = os.path.abspath(os.path.dirname(__file__))
test_dir = os.path.join(root_dir, 'test_cli')
resource_dir = os.path.join(root_dir, '..', 'webscrapbook', 'resources')

class TestServe(unittest.TestCase):
    @mock.patch('sys.stdout', io.StringIO)
    @mock.patch('webscrapbook.cli.server')
    def test_call(self, mock_server):
        cli.cmd_serve({
            'root': test_dir,
            })

        self.assertEqual(mock_server.mock_calls[0], ('serve', (test_dir,), {}))

class TestConfig(unittest.TestCase):
    def tearDown(self):
        try:
            shutil.rmtree(os.path.join(test_dir, WSB_DIR))
        except NotADirectoryError:
            os.remove(os.path.join(test_dir, WSB_DIR))
        except FileNotFoundError:
            pass

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_book_create(self, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': False,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            with open(os.path.join(resource_dir, 'config.ini')) as f2:
                self.assertTrue(f1.read(), f2.read())

        self.assertNotEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_book_create2(self, mock_stdout):
        """No overwrite."""
        os.makedirs(os.path.join(test_dir, WSB_DIR), exist_ok=True)
        with open(os.path.join(test_dir, WSB_DIR, 'config.ini'), 'w') as f:
            f.write('dummy')

        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': False,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            self.assertTrue(f1.read(), 'dummy')

        self.assertEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.util.launch')
    def test_book_edit(self, mock_launch, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': False,
            'edit': True,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            with open(os.path.join(resource_dir, 'config.ini')) as f2:
                self.assertTrue(f1.read(), f2.read())

        mock_launch.assert_called_once_with(os.path.join(test_dir, WSB_DIR, 'config.ini'))

        self.assertNotEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.util.launch')
    def test_book_edit2(self, mock_launch, mock_stdout):
        os.makedirs(os.path.join(test_dir, WSB_DIR), exist_ok=True)
        with open(os.path.join(test_dir, WSB_DIR, 'config.ini'), 'w') as f:
            f.write('foo')

        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': False,
            'edit': True,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            self.assertTrue(f1.read(), 'foo')

        mock_launch.assert_called_once_with(os.path.join(test_dir, WSB_DIR, 'config.ini'))

        self.assertEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_book_all(self, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': True,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            with open(os.path.join(resource_dir, 'config.ini')) as f2:
                self.assertTrue(f1.read(), f2.read())

        with open(os.path.join(test_dir, WSB_DIR, 'serve.py')) as f1:
            with open(os.path.join(resource_dir, 'serve.py')) as f2:
                self.assertTrue(f1.read(), f2.read())

        with open(os.path.join(test_dir, WSB_DIR, 'app.py')) as f1:
            with open(os.path.join(resource_dir, 'app.py')) as f2:
                self.assertTrue(f1.read(), f2.read())

        self.assertNotEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_book_all2(self, mock_stdout):
        """No overwrite."""
        os.makedirs(os.path.join(test_dir, WSB_DIR), exist_ok=True)
        with open(os.path.join(test_dir, WSB_DIR, 'config.ini'), 'w') as f:
            f.write('dummy1')
        with open(os.path.join(test_dir, WSB_DIR, 'serve.py'), 'w') as f:
            f.write('dummy2')
        with open(os.path.join(test_dir, WSB_DIR, 'app.py'), 'w') as f:
            f.write('dummy3')

        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': True,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            self.assertTrue(f1.read(), 'dummy1')

        with open(os.path.join(test_dir, WSB_DIR, 'serve.py')) as f1:
            self.assertTrue(f1.read(), 'dummy2')

        with open(os.path.join(test_dir, WSB_DIR, 'app.py')) as f1:
            self.assertTrue(f1.read(), 'dummy3')

        self.assertEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.cli.WSB_USER_CONFIG', os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))
    def test_user_create(self, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': True,
            'all': False,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'userconfig.ini')) as f1:
            with open(os.path.join(resource_dir, 'config.ini')) as f2:
                self.assertTrue(f1.read(), f2.read())

        self.assertNotEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.cli.WSB_USER_CONFIG', os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))
    def test_user_create2(self, mock_stdout):
        os.makedirs(os.path.join(test_dir, WSB_DIR), exist_ok=True)
        with open(os.path.join(test_dir, WSB_DIR, 'userconfig.ini'), 'w') as f:
            f.write('dummy')

        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': True,
            'all': False,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'userconfig.ini')) as f1:
            self.assertTrue(f1.read(), 'dummy')

        self.assertEqual(mock_stdout.getvalue(), '')

    @mock.patch('webscrapbook.cli.WSB_USER_CONFIG', os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))
    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.util.launch')
    def test_user_edit(self, mock_launch, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': True,
            'all': False,
            'edit': True,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'userconfig.ini')) as f1:
            with open(os.path.join(resource_dir, 'config.ini')) as f2:
                self.assertTrue(f1.read(), f2.read())

        mock_launch.assert_called_once_with(os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))

        self.assertNotEqual(mock_stdout.getvalue(), '')

    @mock.patch('webscrapbook.cli.WSB_USER_CONFIG', os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))
    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.util.launch')
    def test_user_edit2(self, mock_launch, mock_stdout):
        os.makedirs(os.path.join(test_dir, WSB_DIR), exist_ok=True)
        with open(os.path.join(test_dir, WSB_DIR, 'userconfig.ini'), 'w') as f:
            f.write('dummy')

        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': True,
            'all': False,
            'edit': True,
            'name': None,
            })

        mock_launch.assert_called_once_with(os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))

        self.assertEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stderr', new_callable=io.StringIO)
    @mock.patch('sys.exit')
    def test_edit(self, mock_exit, mock_stderr):
        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': False,
            'all': False,
            'edit': True,
            'name': None,
            })

        mock_exit.assert_called_once_with(1)
        self.assertNotEqual(mock_stderr.getvalue(), '')

    @mock.patch('sys.stderr', new_callable=io.StringIO)
    @mock.patch('sys.exit')
    def test_all(self, mock_exit, mock_stderr):
        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': False,
            'all': True,
            'edit': False,
            'name': None,
            })

        mock_exit.assert_called_once_with(1)
        self.assertNotEqual(mock_stderr.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.config.getname', return_value='dummy')
    def test_name_call(self, mock_getname, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': False,
            'all': False,
            'edit': False,
            'name': 'app.name',
            })

        mock_getname.assert_called_once_with('app.name')
        self.assertEqual(mock_stdout.getvalue(), 'dummy\n')

    @mock.patch('sys.exit')
    @mock.patch('sys.stderr', new_callable=io.StringIO)
    @mock.patch('webscrapbook.config.getname', return_value=None)
    def test_name_call2(self, mock_getname, mock_stderr, mock_exit):
        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': False,
            'all': False,
            'edit': False,
            'name': 'unknown.config',
            })

        mock_getname.assert_called_once_with('unknown.config')
        self.assertNotEqual(mock_stderr.getvalue(), '')
        mock_exit.assert_called_once_with(1)

    @mock.patch('webscrapbook.config.dump')
    def test_dump(self, mock_dump):
        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': False,
            'all': False,
            'edit': False,
            'name': None,
            })

        mock_dump.assert_called_once_with(sys.stdout)

class TestEncrypt(unittest.TestCase):
    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.util.encrypt', return_value='dummy_hash')
    def test_call(self, mock_encrypt, mock_stdout):
        cli.cmd_encrypt({
            'password': '1234',
            'salt': 'mysalt',
            'method': 'sha256',
            })

        mock_encrypt.assert_called_once_with('1234', salt='mysalt', method='sha256')
        self.assertEqual(mock_stdout.getvalue(), 'dummy_hash\n')

class TestHelp(unittest.TestCase):
    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_call(self, mock_stdout):
        cli.cmd_help({
            'topic': 'config',
            })

        with open(os.path.join(resource_dir, 'config.md')) as f:
            self.assertEqual(mock_stdout.getvalue(), f.read() + '\n')

class TestView(unittest.TestCase):
    @mock.patch('webscrapbook.cli.view_archive_files')
    def test_call(self, mock_view):
        cli.cmd_view({
            'root': 'config',
            'files': [test_dir],
            })

        mock_view.assert_called_once_with([test_dir])

class TestHelpers(unittest.TestCase):
    def tearDown(self):
        try:
            shutil.rmtree(os.path.join(test_dir, 'temp'))
        except NotADirectoryError:
            os.remove(os.path.join(test_dir, 'temp'))
        except FileNotFoundError:
            pass

    @mock.patch('tempfile.gettempdir', lambda: os.path.join(test_dir, 'temp', 'tempdir'))
    @mock.patch('webbrowser.get', new_callable=mock.MagicMock())
    def test_view_archive_files(self, mock_browser):
        testfile1 = os.path.join(test_dir, 'temp', 'test1.htz')
        testfile2 = os.path.join(test_dir, 'temp', 'test2.maff')
        testfile3 = os.path.join(test_dir, 'temp', 'test3.maff')
        testfile4 = os.path.join(test_dir, 'temp', 'test4.zip')

        os.makedirs(os.path.join(test_dir, 'temp', 'tempdir'), exist_ok=True)
        with zipfile.ZipFile(testfile1, 'w') as zh:
            zh.writestr('index.html', 'abc 測試')
        with zipfile.ZipFile(testfile2, 'w') as zh:
            zh.writestr('123456/index.html', 'abc 測試')
        with zipfile.ZipFile(testfile3, 'w') as zh:
            zh.writestr('abc/index.html', 'abc 測試')
            zh.writestr('def/index.html', 'abc 測試')
        with zipfile.ZipFile(testfile4, 'w') as zh:
            zh.writestr('index.html', 'abc 測試')

        # test jar, no temp dir should be generated
        with mock.patch('webscrapbook.cli.config', {
                'browser': {
                    'command': '',
                    'cache_prefix': 'webscrapbook.',
                    'cache_expire': 1000,
                    'use_jar': True,
                    }
                }):
            cli.view_archive_files([testfile1, testfile2, testfile3, testfile4])

        self.assertEqual(len(mock_browser.mock_calls), 5)
        self.assertEqual(mock_browser.mock_calls[1][1][0],
            r'jar:file:{}!/index.html'.format(pathname2url(os.path.normcase(testfile1)))
            )
        self.assertEqual(mock_browser.mock_calls[2][1][0],
            r'jar:file:{}!/123456/index.html'.format(pathname2url(os.path.normcase(testfile2)))
            )
        self.assertEqual(mock_browser.mock_calls[3][1][0],
            r'jar:file:{}!/abc/index.html'.format(pathname2url(os.path.normcase(testfile3)))
            )
        self.assertEqual(mock_browser.mock_calls[4][1][0],
            r'jar:file:{}!/def/index.html'.format(pathname2url(os.path.normcase(testfile3)))
            )
        self.assertEqual(len(os.listdir(os.path.join(test_dir, 'temp', 'tempdir'))), 0)

        # test simple view
        mock_browser.reset_mock()
        with mock.patch('webscrapbook.cli.config', {
                'browser': {
                    'command': '',
                    'cache_prefix': 'webscrapbook.',
                    'cache_expire': 1000,
                    'use_jar': False,
                    }
                }):
            cli.view_archive_files([testfile1, testfile2, testfile3, testfile4])

        mock_browser.assert_called_once_with(None)
        self.assertEqual(len(mock_browser.mock_calls), 5)
        self.assertRegex(mock_browser.mock_calls[1][1][0],
            r'^file:{}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/index\.html$'.format(
                re.escape(pathname2url(os.path.join(test_dir, 'temp', 'tempdir')))
                ))
        self.assertRegex(mock_browser.mock_calls[2][1][0],
            r'^file:{}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/123456/index\.html$'.format(
                re.escape(pathname2url(os.path.join(test_dir, 'temp', 'tempdir')))
                ))
        self.assertRegex(mock_browser.mock_calls[3][1][0],
            r'^file:{}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/abc/index\.html$'.format(
                re.escape(pathname2url(os.path.join(test_dir, 'temp', 'tempdir')))
                ))
        self.assertRegex(mock_browser.mock_calls[4][1][0],
            r'^file:{}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/def/index\.html$'.format(
                re.escape(pathname2url(os.path.join(test_dir, 'temp', 'tempdir')))
                ))
        self.assertEqual(len(os.listdir(os.path.join(test_dir, 'temp', 'tempdir'))), 3)

        # test browser command
        # test if cache is used for same archive
        mock_browser.reset_mock()
        with mock.patch('webscrapbook.cli.config', {
                'browser': {
                    'command': '/path/to/firefox',
                    'cache_prefix': 'webscrapbook.',
                    'cache_expire': 1000,
                    'use_jar': False,
                    }
                }):
            cli.view_archive_files([testfile1, testfile2, testfile3, testfile4])

        mock_browser.assert_called_once_with('/path/to/firefox')
        self.assertEqual(len(mock_browser.mock_calls), 5)
        self.assertRegex(mock_browser.mock_calls[1][1][0],
            r'^file:{}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/index\.html$'.format(
                re.escape(pathname2url(os.path.join(test_dir, 'temp', 'tempdir')))
                ))
        self.assertRegex(mock_browser.mock_calls[2][1][0],
            r'^file:{}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/123456/index\.html$'.format(
                re.escape(pathname2url(os.path.join(test_dir, 'temp', 'tempdir')))
                ))
        self.assertRegex(mock_browser.mock_calls[3][1][0],
            r'^file:{}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/abc/index\.html$'.format(
                re.escape(pathname2url(os.path.join(test_dir, 'temp', 'tempdir')))
                ))
        self.assertRegex(mock_browser.mock_calls[4][1][0],
            r'^file:{}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/def/index\.html$'.format(
                re.escape(pathname2url(os.path.join(test_dir, 'temp', 'tempdir')))
                ))
        self.assertEqual(len(os.listdir(os.path.join(test_dir, 'temp', 'tempdir'))), 3)

        # test auto clearance of stale caches
        mock_browser.reset_mock()
        with mock.patch('webscrapbook.cli.config', {
                'browser': {
                    'command': '',
                    'cache_prefix': 'webscrapbook.',
                    'cache_expire': -1,
                    'use_jar': False,
                    }
                }):
            with mock.patch('shutil.rmtree') as mock_rmtree:
                cli.view_archive_files([])
                self.assertEqual(len(mock_rmtree.call_args_list), 3)
                self.assertRegex(
                    mock_rmtree.call_args_list[0][0][0].path,
                    r'^{}[0-9a-z]*_[0-9a-z_]*$'.format(re.escape(os.path.join(test_dir, 'temp', 'tempdir', 'webscrapbook.')))
                    )
                self.assertRegex(
                    mock_rmtree.call_args_list[1][0][0].path,
                    r'^{}[0-9a-z]*_[0-9a-z_]*$'.format(re.escape(os.path.join(test_dir, 'temp', 'tempdir', 'webscrapbook.')))
                    )
                self.assertRegex(
                    mock_rmtree.call_args_list[2][0][0].path,
                    r'^{}[0-9a-z]*_[0-9a-z_]*$'.format(re.escape(os.path.join(test_dir, 'temp', 'tempdir', 'webscrapbook.')))
                    )

if __name__ == '__main__':
    unittest.main()



from unittest import mock
import unittest
import sys
import os
import shutil
import io
import time
import zipfile
import collections
from webscrapbook import util
from webscrapbook.util import frozendict

root_dir = os.path.abspath(os.path.dirname(__file__))

class TestUtils(unittest.TestCase):
    def test_frozendict(self):
        dict_ = {'a': 1, 'b': 2, 'c': 3}
        frozendict_ = frozendict(dict_)

        self.assertTrue(isinstance(frozendict_, collections.abc.Hashable))
        self.assertTrue(isinstance(frozendict_, collections.abc.Mapping))
        self.assertFalse(isinstance(frozendict_, collections.abc.MutableMapping))

        self.assertEqual(eval(repr(frozendict_)), frozendict_)
        self.assertRegex(repr(frozendict_), r'^frozendict\([^)]*\)$')

        self.assertTrue(frozendict_ == dict_)
        self.assertTrue('a' in frozendict_)
        self.assertEqual(set(frozendict_), {'a', 'b', 'c'})
        self.assertEqual(list(reversed(frozendict_)), list(frozendict_)[::-1])

        with self.assertRaises(TypeError):
            frozendict_['a'] = 2
        with self.assertRaises(TypeError):
            del frozendict_['a']

        frozendict2 = frozendict_.copy()
        self.assertEqual(frozendict_, frozendict2)
        self.assertIsNot(frozendict_, frozendict2)

    def test_make_hashable(self):
        self.assertEqual(
            type(util.make_hashable({1, 2, 3})),
            frozenset
            )

        self.assertEqual(
            type(util.make_hashable(['foo', 'bar', 'baz'])),
            tuple
            )

        self.assertEqual(
            type(util.make_hashable({'a': 123, 'b': 456, 'c': 789})),
            frozendict
            )
            
        self.assertEqual(
            set(util.make_hashable([{'a': 123, 'b': 456}, [1, 2, 3]])),
            {(1, 2, 3), frozendict({'a': 123, 'b': 456})}
            )

    def test_is_nullhost(self):
        self.assertTrue(util.is_nullhost('0.0.0.0'))
        self.assertFalse(util.is_nullhost('127.0.0.1'))
        self.assertFalse(util.is_localhost('192.168.0.1'))
        self.assertTrue(util.is_nullhost('::'))
        self.assertFalse(util.is_nullhost('::1'))
        self.assertTrue(util.is_nullhost('0::0'))
        self.assertTrue(util.is_nullhost('0000::0000'))
        self.assertTrue(util.is_nullhost('0000:0000::0000'))
        self.assertTrue(util.is_nullhost('0:0:0:0:0:0:0:0'))
        self.assertTrue(util.is_nullhost('0000:0000:0000:0000:0000:0000:0000:0000'))
        self.assertFalse(util.is_nullhost('wtf'))

    def test_is_localhost(self):
        self.assertFalse(util.is_localhost('0.0.0.0'))
        self.assertTrue(util.is_localhost('127.0.0.1'))
        self.assertFalse(util.is_localhost('192.168.0.1'))
        self.assertFalse(util.is_localhost('::'))
        self.assertTrue(util.is_localhost('::1'))
        self.assertTrue(util.is_localhost('0:0:0:0:0:0:0:1'))
        self.assertTrue(util.is_localhost('0000:0000:0000:0000:0000:0000:0000:0001'))
        self.assertFalse(util.is_localhost('wtf'))

    def test_get_breadcrumbs(self):
        self.assertEqual(list(util.get_breadcrumbs('/path/to/directory/')), [
            ('.', '/', '/', False),
            ('path', '/path/', '/', False),
            ('to', '/path/to/', '/', False),
            ('directory', '/path/to/directory/', '/', True)
            ])

        self.assertEqual(list(util.get_breadcrumbs('/path/to/archive.ext!/')), [
            ('.', '/', '/', False),
            ('path', '/path/', '/', False),
            ('to', '/path/to/', '/', False),
            ('archive.ext!', '/path/to/archive.ext!/', '/', True),
            ])

        # subarchivepath
        self.assertEqual(list(util.get_breadcrumbs('/path/to/archive.ext!/', subarchivepath='')), [
            ('.', '/', '/', False),
            ('path', '/path/', '/', False),
            ('to', '/path/to/', '/', False),
            ('archive.ext', '/path/to/archive.ext!/', '!/', True),
            ])

        self.assertEqual(list(util.get_breadcrumbs('/path/to/archive.ext!/subdir', subarchivepath='subdir')), [
            ('.', '/', '/', False),
            ('path', '/path/', '/', False),
            ('to', '/path/to/', '/', False),
            ('archive.ext', '/path/to/archive.ext!/', '!/', False),
            ('subdir', '/path/to/archive.ext!/subdir/', '/', True),
            ])

        # base
        self.assertEqual(list(util.get_breadcrumbs('/path/to/directory/', base='/wsb')), [
            ('.', '/wsb/', '/', False),
            ('path', '/wsb/path/', '/', False),
            ('to', '/wsb/path/to/', '/', False),
            ('directory', '/wsb/path/to/directory/', '/', True),
            ])

        self.assertEqual(list(util.get_breadcrumbs('/path/to/directory/', base='/wsb/')), [
            ('.', '/wsb/', '/', False),
            ('path', '/wsb/path/', '/', False),
            ('to', '/wsb/path/to/', '/', False),
            ('directory', '/wsb/path/to/directory/', '/', True),
            ])

        # topname
        self.assertEqual(list(util.get_breadcrumbs('/path/to/directory/', topname='MyWsb')), [
            ('MyWsb', '/', '/', False),
            ('path', '/path/', '/', False),
            ('to', '/path/to/', '/', False),
            ('directory', '/path/to/directory/', '/', True)
            ])

    def test_checksum(self):
        self.assertEqual(
            util.checksum(os.path.join(root_dir, 'test_util', 'checksum', 'checksum.txt')),
            'da39a3ee5e6b4b0d3255bfef95601890afd80709'
            )

        self.assertEqual(
            util.checksum(os.path.join(root_dir, 'test_util', 'checksum', 'checksum.txt'), method='sha256'),
            'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
            )

    def test_file_info(self):
        entry = os.path.join(root_dir, 'test_util', 'file_info', 'nonexist.file')
        self.assertEqual(
            util.file_info(entry),
            ('nonexist.file', None, None, None)
            )

        entry = os.path.join(root_dir, 'test_util', 'file_info', 'file.txt')
        self.assertEqual(
            util.file_info(entry),
            ('file.txt', 'file', 3, os.stat(entry).st_mtime)
            )

        entry = os.path.join(root_dir, 'test_util', 'file_info', 'folder')
        self.assertEqual(
            util.file_info(entry),
            ('folder', 'dir', None, os.stat(entry).st_mtime)
            )

    def test_listdir(self):
        entry = os.path.join(root_dir, 'test_util', 'listdir')
        self.assertEqual(set(util.listdir(entry)), {
            ('file.txt', 'file', 3, os.stat(os.path.join(entry, 'file.txt')).st_mtime),
            ('folder', 'dir', None, os.stat(os.path.join(entry, 'folder')).st_mtime),
            })
        self.assertEqual(set(util.listdir(entry, recursive=True)), {
            ('file.txt', 'file', 3, os.stat(os.path.join(entry, 'file.txt')).st_mtime),
            ('folder', 'dir', None, os.stat(os.path.join(entry, 'folder')).st_mtime),
            ('folder/.gitkeep', 'file', 0, os.stat(os.path.join(entry, 'folder', '.gitkeep')).st_mtime),
            })

    def test_format_filesize(self):
        self.assertEqual(util.format_filesize(0), '0 B')
        self.assertEqual(util.format_filesize(3), '3 B')
        self.assertEqual(util.format_filesize(1000), '1000 B')
        self.assertEqual(util.format_filesize(1024), '1.0 KB')
        self.assertEqual(util.format_filesize(1080), '1.1 KB')
        self.assertEqual(util.format_filesize(10000), '9.8 KB')
        self.assertEqual(util.format_filesize(10240), '10 KB')
        self.assertEqual(util.format_filesize(20480), '20 KB')
        self.assertEqual(util.format_filesize(1048576), '1.0 MB')
        self.assertEqual(util.format_filesize(2621440), '2.5 MB')
        self.assertEqual(util.format_filesize(10485760), '10 MB')
        self.assertEqual(util.format_filesize(1073741824), '1.0 GB')
        self.assertEqual(util.format_filesize(10737418240), '10 GB')
        self.assertEqual(util.format_filesize(1e14), '91 TB')
        self.assertEqual(util.format_filesize(1e28), '8272 YB')

        self.assertEqual(util.format_filesize(0, si=True), '0 B')
        self.assertEqual(util.format_filesize(3, si=True), '3 B')
        self.assertEqual(util.format_filesize(1000, si=True), '1.0 kB')
        self.assertEqual(util.format_filesize(1024, si=True), '1.0 kB')
        self.assertEqual(util.format_filesize(1080, si=True), '1.1 kB')
        self.assertEqual(util.format_filesize(10000, si=True), '10 kB')
        self.assertEqual(util.format_filesize(10240, si=True), '10 kB')
        self.assertEqual(util.format_filesize(20480, si=True), '20 kB')
        self.assertEqual(util.format_filesize(1048576, si=True), '1.0 MB')
        self.assertEqual(util.format_filesize(2621440, si=True), '2.6 MB')
        self.assertEqual(util.format_filesize(10485760, si=True), '10 MB')
        self.assertEqual(util.format_filesize(1073741824, si=True), '1.1 GB')
        self.assertEqual(util.format_filesize(10737418240, si=True), '11 GB')
        self.assertEqual(util.format_filesize(1e14, si=True), '100 TB')
        self.assertEqual(util.format_filesize(1e28, si=True), '10000 YB')

    def test_zip_file_info(self):
        zip_filename = os.path.join(root_dir, 'test_util', 'zipfile.zip')
        try:
            with zipfile.ZipFile(zip_filename, 'w') as zh:
                zh.writestr(zipfile.ZipInfo('file.txt', (1987, 1, 1, 0, 0, 0)), '123456')
                zh.writestr(zipfile.ZipInfo('folder/', (1988, 1, 1, 0, 0, 0)), '')
                zh.writestr(zipfile.ZipInfo('folder/.gitkeep', (1989, 1, 1, 0, 0, 0)), '123')
                zh.writestr(zipfile.ZipInfo('implicit_folder/.gitkeep', (1990, 1, 1, 0, 0, 0)), '1234')

            self.assertEqual(
                util.zip_file_info(zip_filename, 'file.txt'),
                ('file.txt', 'file', 6, 536428800)
                )

            self.assertEqual(
                util.zip_file_info(zip_filename, 'folder'),
                ('folder', 'dir', None, 567964800)
                )

            self.assertEqual(
                util.zip_file_info(zip_filename, 'folder/'),
                ('folder', 'dir', None, 567964800)
                )

            self.assertEqual(
                util.zip_file_info(zip_filename, 'folder/.gitkeep'),
                ('.gitkeep', 'file', 3, 599587200)
                )

            self.assertEqual(
                util.zip_file_info(zip_filename, 'implicit_folder'),
                ('implicit_folder', None, None, None)
                )

            self.assertEqual(
                util.zip_file_info(zip_filename, 'implicit_folder/'),
                ('implicit_folder', None, None, None)
                )

            self.assertEqual(
                util.zip_file_info(zip_filename, 'implicit_folder', check_implicit_dir=True),
                ('implicit_folder', 'dir', None, None)
                )

            self.assertEqual(
                util.zip_file_info(zip_filename, 'implicit_folder/', check_implicit_dir=True),
                ('implicit_folder', 'dir', None, None)
                )

            self.assertEqual(
                util.zip_file_info(zip_filename, 'implicit_folder/.gitkeep'),
                ('.gitkeep', 'file', 4, 631123200)
                )

            self.assertEqual(
                util.zip_file_info(zip_filename, 'nonexist'),
                ('nonexist', None, None, None)
                )

            self.assertEqual(
                util.zip_file_info(zip_filename, 'nonexist/'),
                ('nonexist', None, None, None)
                )
        finally:
            try:
                os.remove(zip_filename)
            except FileNotFoundError:
                pass

    def test_zip_listdir(self):
        zip_filename = os.path.join(root_dir, 'test_util', 'zipfile.zip')
        try:
            with zipfile.ZipFile(zip_filename, 'w') as zh:
                zh.writestr(zipfile.ZipInfo('file.txt', (1987, 1, 1, 0, 0, 0)), '123456')
                zh.writestr(zipfile.ZipInfo('folder/', (1988, 1, 1, 0, 0, 0)), '')
                zh.writestr(zipfile.ZipInfo('folder/.gitkeep', (1989, 1, 1, 0, 0, 0)), '123')
                zh.writestr(zipfile.ZipInfo('implicit_folder/.gitkeep', (1990, 1, 1, 0, 0, 0)), '1234')

            self.assertEqual(set(util.zip_listdir(zip_filename, '')), {
                ('folder', 'dir', None, 567964800),
                ('implicit_folder', 'dir', None, None),
                ('file.txt', 'file', 6, 536428800),
                })

            self.assertEqual(set(util.zip_listdir(zip_filename, '/')), {
                ('folder', 'dir', None, 567964800),
                ('implicit_folder', 'dir', None, None),
                ('file.txt', 'file', 6, 536428800),
                })

            self.assertEqual(set(util.zip_listdir(zip_filename, '', recursive=True)), {
                ('folder', 'dir', None, 567964800),
                ('folder/.gitkeep', 'file', 3, 599587200),
                ('implicit_folder', 'dir', None, None),
                ('implicit_folder/.gitkeep', 'file', 4, 631123200),
                ('file.txt', 'file', 6, 536428800),
                })

            self.assertEqual(set(util.zip_listdir(zip_filename, 'folder')), {
                ('.gitkeep', 'file', 3, 599587200)
                })

            self.assertEqual(set(util.zip_listdir(zip_filename, 'folder/')), {
                ('.gitkeep', 'file', 3, 599587200)
                })

            self.assertEqual(set(util.zip_listdir(zip_filename, 'implicit_folder')), {
                ('.gitkeep', 'file', 4, 631123200)
                })

            self.assertEqual(set(util.zip_listdir(zip_filename, 'implicit_folder/')), {
                ('.gitkeep', 'file', 4, 631123200)
                })

            with self.assertRaises(util.ZipDirNotFoundError):
                set(util.zip_listdir(zip_filename, 'nonexist'))

            with self.assertRaises(util.ZipDirNotFoundError):
                set(util.zip_listdir(zip_filename, 'nonexist/'))

            with self.assertRaises(util.ZipDirNotFoundError):
                set(util.zip_listdir(zip_filename, 'file.txt'))
        finally:
            try:
                os.remove(zip_filename)
            except FileNotFoundError:
                pass

    def test_parse_meta_refresh(self):
        self.assertEqual(
            util.parse_meta_refresh(os.path.join(root_dir, 'test_util', 'parse_meta_refresh', 'refresh1.html')),
            (0, 'target.html')
            )
        self.assertEqual(
            util.parse_meta_refresh(os.path.join(root_dir, 'test_util', 'parse_meta_refresh', 'refresh2.html')),
            (0, 'target.html')
            )

    def test_parse_maff_index_rdf(self):
        maff_filename = os.path.join(root_dir, 'test_util', 'tempfile.maff')
        try:
            with zipfile.ZipFile(maff_filename, 'w') as zh:
                zh.writestr(zipfile.ZipInfo('19870101/index.rdf', (1987, 1, 1, 0, 0, 0)), """<?xml version="1.0"?>
<RDF:RDF xmlns:MAF="http://maf.mozdev.org/metadata/rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:root">
    <MAF:originalurl RDF:resource="http://example.com/"/>
    <MAF:title RDF:resource="Example MAFF"/>
    <MAF:archivetime RDF:resource="Mon, 25 Dec 2017 17:27:46 GMT"/>
    <MAF:indexfilename RDF:resource="index.html"/>
    <MAF:charset RDF:resource="UTF-8"/>
  </RDF:Description>
</RDF:RDF>""")

            with zipfile.ZipFile(maff_filename, 'r') as zh:
                with zh.open('19870101/index.rdf', 'r') as rdf:
                    self.assertEqual(
                        util.parse_maff_index_rdf(rdf),
                        ('Example MAFF', 'http://example.com/', 'Mon, 25 Dec 2017 17:27:46 GMT', 'index.html', 'UTF-8')
                        )
        finally:
            try:
                os.remove(maff_filename)
            except FileNotFoundError:
                pass

    def test_get_maff_pages(self):
        maff_filename = os.path.join(root_dir, 'test_util', 'tempfile.maff')
        try:
            with zipfile.ZipFile(maff_filename, 'w') as zh:
                zh.writestr(zipfile.ZipInfo('webpage1/index.rdf', (1987, 1, 1, 0, 0, 0)), """<?xml version="1.0"?>
<RDF:RDF xmlns:MAF="http://maf.mozdev.org/metadata/rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:root">
    <MAF:originalurl RDF:resource="http://example.com/"/>
    <MAF:title RDF:resource="Example MAFF"/>
    <MAF:archivetime RDF:resource="Mon, 25 Dec 2017 17:27:46 GMT"/>
    <MAF:indexfilename RDF:resource="index.html"/>
    <MAF:charset RDF:resource="UTF-8"/>
  </RDF:Description>
</RDF:RDF>""")
                zh.writestr(zipfile.ZipInfo('webpage2/index.html', (1987, 1, 1, 0, 0, 0)), '')
                zh.writestr(zipfile.ZipInfo('webpage3/index.svg', (1987, 1, 1, 0, 0, 0)), '')

            self.assertEqual(util.get_maff_pages(maff_filename), [
                ('Example MAFF', 'http://example.com/', 'Mon, 25 Dec 2017 17:27:46 GMT', 'webpage1/index.html', 'UTF-8'),
                (None, None, None, 'webpage2/index.html', None),
                (None, None, None, 'webpage3/index.svg', None),
                ])
        finally:
            try:
                os.remove(maff_filename)
            except FileNotFoundError:
                pass

    @mock.patch('sys.stderr', io.StringIO())
    def test_encrypt(self):
        self.assertEqual(util.encrypt('1234', 'salt', 'plain'),
            '1234salt'
            )
        self.assertEqual(util.encrypt('1234', 'salt', 'md5'),
            '1fadcf6eb4345975be993f237c51d426'
            )
        self.assertEqual(util.encrypt('1234', 'salt', 'sha1'),
            '40c95464b7eacddb5572af5468ffb1cdb5b13f35'
            )
        self.assertEqual(util.encrypt('1234', 'salt', 'sha256'),
            '4b3bed8af7b7612e8c1e25f63ba24496f5b16b2df44efb2db7ce3cb24b7e96f7'
            )
        self.assertEqual(util.encrypt('1234', 'salt', 'unknown'),
            '1234salt'
            )

class TestTokenHandler(unittest.TestCase):
    def setUp(self):
        self.test_dir = os.path.join(os.path.dirname(__file__), 'test_util', 'token_handler')
        os.makedirs(self.test_dir, exist_ok=True)

    def tearDown(self):
        try:
            shutil.rmtree(self.test_dir)
        except NotADirectoryError:
            os.remove(self.test_dir)
        except FileNotFoundError:
            pass

    @mock.patch('webscrapbook.util.TokenHandler.check_delete_expire')
    @mock.patch('webscrapbook.util.TokenHandler.DEFAULT_EXPIRY', 10)
    def test_acquire1(self, mock_check):
        now = time.time()
        expected_expire_time = int(now) + 10

        handler = util.TokenHandler(self.test_dir)
        token = handler.acquire()
        token_file = os.path.join(self.test_dir, token)

        self.assertTrue(os.path.isfile(token_file))
        with open(token_file, 'r', encoding='UTF-8') as f:
            self.assertAlmostEqual(int(f.read()), expected_expire_time, delta=1)
        self.assertAlmostEqual(mock_check.call_args[0][0], now, delta=1)

    @mock.patch('webscrapbook.util.TokenHandler.check_delete_expire')
    @mock.patch('webscrapbook.util.TokenHandler.DEFAULT_EXPIRY', 30)
    def test_acquire2(self, mock_check):
        now = 30000
        expected_expire_time = int(now) + 30

        handler = util.TokenHandler(self.test_dir)
        token = handler.acquire(now)
        token_file = os.path.join(self.test_dir, token)

        self.assertTrue(os.path.isfile(token_file))
        with open(token_file, 'r', encoding='UTF-8') as f:
            self.assertEqual(int(f.read()), expected_expire_time)
        self.assertEqual(mock_check.call_args[0][0], now)

    def test_validate1(self):
        token = 'sampleToken'
        token_time = int(time.time()) + 3

        token_file = os.path.join(self.test_dir, token)
        with open(token_file, 'w', encoding='UTF-8') as f:
            f.write(str(token_time))

        handler = util.TokenHandler(self.test_dir)
        self.assertTrue(handler.validate(token))

    def test_validate2(self):
        token = 'sampleToken'
        token_time = int(time.time()) - 3

        token_file = os.path.join(self.test_dir, token)
        with open(token_file, 'w', encoding='UTF-8') as f:
            f.write(str(token_time))

        handler = util.TokenHandler(self.test_dir)
        self.assertFalse(handler.validate(token))

    def test_validate3(self):
        token = 'sampleToken'
        now = 30000
        token_time = 30001

        token_file = os.path.join(self.test_dir, token)
        with open(token_file, 'w', encoding='UTF-8') as f:
            f.write(str(token_time))

        handler = util.TokenHandler(self.test_dir)
        self.assertTrue(handler.validate(token, now))

    def test_validate4(self):
        token = 'sampleToken'
        now = 30000
        token_time = 29999

        token_file = os.path.join(self.test_dir, token)
        with open(token_file, 'w', encoding='UTF-8') as f:
            f.write(str(token_time))

        handler = util.TokenHandler(self.test_dir)
        self.assertFalse(handler.validate(token, now))

    def test_delete(self):
        token = 'sampleToken'

        token_file = os.path.join(self.test_dir, token)
        with open(token_file, 'w', encoding='UTF-8') as f:
            f.write(str(32768))

        handler = util.TokenHandler(self.test_dir)
        handler.delete(token)
        self.assertFalse(os.path.exists(token_file))

    def test_delete_expire1(self):
        now = int(time.time())

        with open(os.path.join(self.test_dir, 'sampleToken1'), 'w', encoding='UTF-8') as f:
            f.write(str(now - 100))
        with open(os.path.join(self.test_dir, 'sampleToken2'), 'w', encoding='UTF-8') as f:
            f.write(str(now - 10))
        with open(os.path.join(self.test_dir, 'sampleToken3'), 'w', encoding='UTF-8') as f:
            f.write(str(now + 10))
        with open(os.path.join(self.test_dir, 'sampleToken4'), 'w', encoding='UTF-8') as f:
            f.write(str(now + 100))

        handler = util.TokenHandler(self.test_dir)
        handler.delete_expire()

        self.assertFalse(os.path.exists(os.path.join(self.test_dir, 'sampleToken1')))
        self.assertFalse(os.path.exists(os.path.join(self.test_dir, 'sampleToken2')))
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, 'sampleToken3')))
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, 'sampleToken4')))

    def test_delete_expire2(self):
        now = 30000

        with open(os.path.join(self.test_dir, 'sampleToken1'), 'w', encoding='UTF-8') as f:
            f.write(str(29000))
        with open(os.path.join(self.test_dir, 'sampleToken2'), 'w', encoding='UTF-8') as f:
            f.write(str(29100))
        with open(os.path.join(self.test_dir, 'sampleToken3'), 'w', encoding='UTF-8') as f:
            f.write(str(30100))
        with open(os.path.join(self.test_dir, 'sampleToken4'), 'w', encoding='UTF-8') as f:
            f.write(str(30500))

        handler = util.TokenHandler(self.test_dir)
        handler.delete_expire(now)

        self.assertFalse(os.path.exists(os.path.join(self.test_dir, 'sampleToken1')))
        self.assertFalse(os.path.exists(os.path.join(self.test_dir, 'sampleToken2')))
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, 'sampleToken3')))
        self.assertTrue(os.path.exists(os.path.join(self.test_dir, 'sampleToken4')))

    @mock.patch('webscrapbook.util.TokenHandler.delete_expire')
    def test_check_delete_expire1(self, mock_delete):
        now = int(time.time())

        handler = util.TokenHandler(self.test_dir)
        self.assertEqual(handler.last_purge, 0)

        handler.check_delete_expire()
        self.assertAlmostEqual(mock_delete.call_args[0][0], now, delta=1)
        self.assertAlmostEqual(handler.last_purge, now, delta=1)

    @mock.patch('webscrapbook.util.TokenHandler.delete_expire')
    @mock.patch('webscrapbook.util.TokenHandler.PURGE_INTERVAL', 1000)
    def test_check_delete_expire2(self, mock_delete):
        now = int(time.time())

        handler = util.TokenHandler(self.test_dir)
        handler.last_purge = now - 1100

        handler.check_delete_expire()
        self.assertAlmostEqual(mock_delete.call_args[0][0], now, delta=1)
        self.assertAlmostEqual(handler.last_purge, now, delta=1)

    @mock.patch('webscrapbook.util.TokenHandler.delete_expire')
    @mock.patch('webscrapbook.util.TokenHandler.PURGE_INTERVAL', 1000)
    def test_check_delete_expire3(self, mock_delete):
        now = int(time.time())

        handler = util.TokenHandler(self.test_dir)
        handler.last_purge = now - 900

        handler.check_delete_expire()
        mock_delete.assert_not_called()
        self.assertEqual(handler.last_purge, now - 900)

    @mock.patch('webscrapbook.util.TokenHandler.delete_expire')
    @mock.patch('webscrapbook.util.TokenHandler.PURGE_INTERVAL', 1000)
    def test_check_delete_expire4(self, mock_delete):
        now = 40000

        handler = util.TokenHandler(self.test_dir)
        handler.last_purge = now - 1100

        handler.check_delete_expire(now)
        self.assertAlmostEqual(mock_delete.call_args[0][0], now, delta=1)
        self.assertAlmostEqual(handler.last_purge, now, delta=1)

    @mock.patch('webscrapbook.util.TokenHandler.delete_expire')
    @mock.patch('webscrapbook.util.TokenHandler.PURGE_INTERVAL', 1000)
    def test_check_delete_expire5(self, mock_delete):
        now = 40000

        handler = util.TokenHandler(self.test_dir)
        handler.last_purge = now - 900

        handler.check_delete_expire(now)
        mock_delete.assert_not_called()
        self.assertEqual(handler.last_purge, now - 900)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""The WGSI application.
"""
import sys
import os
import traceback
import shutil
import mimetypes
import re
import zipfile
import time
import json
import functools
from urllib.parse import urlsplit, urlunsplit, urljoin, quote, unquote, parse_qs
from pathlib import Path
from zlib import adler32

# dependency
import flask
from flask import request, Response, redirect, abort, render_template, send_from_directory
from flask import current_app
from werkzeug.local import LocalProxy
from werkzeug.middleware.proxy_fix import ProxyFix
from werkzeug.http import is_resource_modified
from werkzeug.http import http_date
from werkzeug.http import parse_options_header, dump_options_header
import jinja2
import commonmark

# this package
from . import *
from . import __version__
from . import Config
from . import util
from ._compat.time import time_ns

# see: https://url.spec.whatwg.org/#percent-encoded-bytes
quote_path = functools.partial(quote, safe=":/[]@!$&'()*+,;=")
quote_path.__doc__ = "Escape reserved chars for the path part of a URL."

bp = flask.Blueprint('default', __name__)
runtime = LocalProxy(lambda: current_app.config['WEBSCRAPBOOK_RUNTIME'])


def static_file(filepath, root=None, mimetype=None):
    """Wrap send_file for customized behaviors.
    """
    response = send_from_directory(root or runtime['root'], filepath, mimetype=mimetype)
    response.headers.set('Accept-Ranges', 'bytes')
    response.headers.set('Cache-Control', 'no-cache')
    return response


def http_response(body='', status=None, headers=None, format=None):
    """Handle formatted response.

    ref: https://jsonapi.org
    """
    if not format:
        mimetype = None

    # expect body to be a JSON-serializable object
    elif format == 'json':
        mimetype = 'application/json'

        body = {
            'success': True,
            'data': body,
            }

        body = json.dumps(body, ensure_ascii=False)

    # expect body to be a generator of text (mostly JSON) data
    elif format == 'sse':
        mimetype = 'text/event-stream'

        def wrapper(gen):
            for data in gen:
                yield "data: " + data + "\n\n"

            yield "event: complete" + "\n"
            yield "data: " + "\n\n"

        body = wrapper(body)

    else:
        return http_error(400, 'Output format "{}" is not supported.'.format(format), format=format)

    return Response(body, status, headers, mimetype=mimetype)


def http_error(status=500, body=None, headers=None, format=None):
    """Handle formatted error response.
    """
    # expect body to be a JSON-serializable object
    if format == 'json':
        mimetype = 'application/json'

        body = {
            'error': {
                'status': status,
                'message': body,
                },
            }

        body = json.dumps(body, ensure_ascii=False)
    else:
        mimetype = None

    if body:
        return Response(body, status, headers, mimetype=mimetype)
    else:
        # @TODO: abort does not support headers and mimetype
        return abort(status)


def get_archive_path(filepath):
    """Parse archive file path and the sub-archive path.

    - Priority:
      entry.zip!/entry1.zip!/ = entry.zip!/entry1.zip! >
      entry.zip!/entry1.zip >
      entry.zip!/ = entry.zip! >
      entry.zip

    Returns:
        a tuple (archivefile, subarchivepath).
    """
    for m in reversed(list(re.finditer(r'!/', filepath, flags=re.I))):
        archivefile = os.path.normpath(os.path.join(runtime['root'], filepath[:m.start(0)].strip('/')))
        if os.path.lexists(archivefile + '!'):
            break

        if zipfile.is_zipfile(archivefile):
            subarchivepath = filepath[m.end(0):].rstrip('/')
            return (archivefile, subarchivepath)

    return (None, None)


def is_local_access():
    """Determine if the client is in same device.
    """
    server_host = request.host.partition(':')[0]
    client_host = request.remote_addr
    return util.is_localhost(server_host) or util.is_localhost(client_host) or server_host == client_host


def get_permission(auth):
    """Get effective permission of the current logged in user.
    """
    try:
        auth_sections = runtime['config']['auth']
    except KeyError:
        return 'all'

    auth = auth or {}
    user = auth.get('username') or ''
    pw = auth.get('password') or ''

    for _, entry in auth_sections.items():
        entry_user = entry.get('user', '')
        entry_pw = entry.get('pw', '')
        entry_pw_salt = entry.get('pw_salt', '')
        entry_pw_type = entry.get('pw_type', '')
        entry_permission = entry.get('permission', 'all')
        if (user == entry_user and
                util.encrypt(pw, entry_pw_salt, entry_pw_type) == entry_pw):
            return entry_permission

    return ''


def verify_authorization(perm, action):
    """Check if authorized or not.
    """
    if perm == 'all':
        return True

    elif perm == 'read':
        if action in {'token', 'lock', 'unlock', 'mkdir', 'save', 'delete', 'move', 'copy'}:
            return False
        else:
            return True

    elif perm == 'view':
        if action in {'view', 'source', 'static'}:
            return True
        else:
            return False

    else:
        return False


def handle_directory_listing(localpath, recursive=False, format=None):
    """List contents in a directory.
    """
    # ensure directory has trailing '/'
    if not request.path.endswith('/'):
        parts = urlsplit(request.url)
        new_url = urlunsplit((
            parts.scheme,
            parts.netloc,
            quote_path(unquote(parts.path)) + '/',
            parts.query,
            parts.fragment,
            ))
        return redirect(new_url)

    if not os.path.exists(localpath):
        return http_error(404, "Directory does not exist.", format=format)

    if not os.access(localpath, os.R_OK):
        return http_error(403, "You do not have permission to view this directory.", format=format)

    # output index
    subentries = util.listdir(localpath, recursive)

    if format == 'sse':
        def gen():
            for entry in subentries:
                data = {
                    'name': entry.name,
                    'type': entry.type,
                    'size': entry.size,
                    'last_modified': entry.last_modified,
                    }

                yield json.dumps(data, ensure_ascii=False)

        return http_response(gen(), format=format)

    elif format == 'json':
        data = []
        for entry in subentries:
            data.append({
                    'name': entry.name,
                    'type': entry.type,
                    'size': entry.size,
                    'last_modified': entry.last_modified,
                    })
        return http_response(data, format=format)

    # headers
    headers = {}
    headers['Cache-Control'] = 'no-store'
    stats = os.stat(localpath)
    headers['Last-Modified'] = http_date(stats.st_mtime)

    body = render_template('index.html',
            sitename=runtime['name'],
            is_local=is_local_access(),
            base=request.script_root,
            path=request.path,
            subarchivepath=None,
            subentries=subentries,
            )

    return http_response(body, format=format, headers=headers)


def handle_zip_directory_listing(zip, archivefile, subarchivepath, recursive=False, format=None):
    """List contents in a directory.
    """
    # ensure directory has trailing '/'
    if not request.path.endswith('/'):
        parts = urlsplit(request.url)
        new_url = urlunsplit((
            parts.scheme,
            parts.netloc,
            quote_path(unquote(parts.path)) + '/',
            parts.query,
            parts.fragment,
            ))
        return redirect(new_url)

    stats = os.lstat(archivefile)
    last_modified = http_date(stats.st_mtime)
    etag = "%s-%s-%s" % (
        stats.st_mtime,
        stats.st_size,
        adler32(archivefile.encode("utf-8")) & 0xFFFFFFFF,
        )

    if not is_resource_modified(request.environ, etag=etag, last_modified=last_modified):
        return http_response(status=304, format=format)

    headers = {
        'Cache-Control': 'no-cache',
        'Last-Modified': last_modified,
        'ETag': etag,
        }

    subentries = util.zip_listdir(zip, subarchivepath, recursive)

    if format == 'sse':
        def gen():
            for entry in subentries:
                data = {
                    'name': entry.name,
                    'type': entry.type,
                    'size': entry.size,
                    'last_modified': entry.last_modified,
                    }

                yield json.dumps(data, ensure_ascii=False)

        return http_response(gen(), headers=headers, format=format)

    elif format == 'json':
        data = []
        for entry in subentries:
            data.append({
                    'name': entry.name,
                    'type': entry.type,
                    'size': entry.size,
                    'last_modified': entry.last_modified,
                    })
        return http_response(data, headers=headers, format=format)

    try:
        body = render_template('index.html',
                sitename=runtime['name'],
                is_local=is_local_access(),
                base=request.script_root,
                path=request.path,
                subarchivepath=subarchivepath,
                subentries=subentries,
                )
        return http_response(body, headers=headers)
    except util.ZipDirNotFoundError:
        return http_error(404, "File does not exist.")


def handle_subarchive_path(archivefile, subarchivepath, mimetype=None, list_directory=True, format=None):
    """Show content of a path in a zip file.
    """
    if not os.access(archivefile, os.R_OK):
        return http_error(403, "You do not have permission to access this file.", format=format)

    try:
        zip = zipfile.ZipFile(archivefile)
    except:
        return http_error(500, "Unable to open the ZIP file.", format=format)

    if format:
        info = util.zip_file_info(zip, subarchivepath)
        data = {
            'name': info.name,
            'type': info.type,
            'size': info.size,
            'last_modified': info.last_modified,
            'mime': mimetype,
            }
        return http_response(data, format=format)

    try:
        # KeyError is raised if subarchivepath does not exist
        info = zip.getinfo(subarchivepath)
    except KeyError:
        # subarchivepath does not exist
        # possibility a missing directory entry?
        if not list_directory:
            return http_error(404)

        return handle_zip_directory_listing(zip, archivefile, subarchivepath)
    else:
        fh = zip.open(info, 'r')

        lm = info.date_time
        lm = int(time.mktime((lm[0], lm[1], lm[2], lm[3], lm[4], lm[5], 0, 0, -1)))
        last_modified = http_date(lm)

        etag = "%s-%s-%s" % (
            lm,
            info.file_size,
            adler32(archivefile.encode("utf-8")) & 0xFFFFFFFF,
            )

        headers = {
            'Accept-Ranges': 'bytes',
            'Cache-Control': 'no-cache',
            'Last-Modified': last_modified,
            'ETag': etag,
            }

        response = Response(fh, headers=headers, mimetype=mimetype)
        response.make_conditional(request.environ, accept_ranges=True, complete_length=info.file_size)
        return response
    finally:
        zip.close()


def handle_archive_viewing(localpath, mimetype):
    """Handle direct visit of HTZ/MAFF file.
    """
    def list_maff_pages(pages):
        """List available web pages in a MAFF file.
        """
        return render_template('maff_index.html',
                sitename=runtime['name'],
                is_local=is_local_access(),
                base=request.script_root,
                path=request.path,
                pages=pages,
                )

    if not os.access(localpath, os.R_OK):
        return http_error(403, "You do not have permission to access this file.")

    if mimetype == "application/html+zip":
        subpath = "index.html"
    else:
        pages = util.get_maff_pages(localpath)

        if len(pages) > 1:
            # multiple index files
            return list_maff_pages(pages)
        elif len(pages) > 0:
            subpath = pages[0].indexfilename
        else:
            # no valid index file found
            return list_maff_pages([])

    parts = urlsplit(request.url)
    new_url = urlunsplit((
        parts.scheme,
        parts.netloc,
        quote_path(unquote(parts.path)) + '!/' + quote_path(subpath),
        parts.query,
        parts.fragment,
        ))
    return redirect(new_url)


def handle_markdown_output(filepath, filename):
    """Output processed markdown.
    """
    stats = os.stat(filename)
    last_modified = http_date(stats.st_mtime)
    etag = "%s-%s-%s" % (
        stats.st_mtime,
        stats.st_size,
        adler32(filename.encode("utf-8")) & 0xFFFFFFFF,
        )

    headers = {
        'Cache-Control': 'no-cache',
        }

    if not is_resource_modified(request.environ, etag=etag, last_modified=last_modified):
        return http_response(status=304, headers=headers)

    headers.update({
        'Last-Modified': last_modified,
        'ETag': etag,
        })

    # output processed content
    with open(filename, 'r', encoding='UTF-8') as f:
        body = f.read()
    
    body = render_template('markdown.html',
            sitename=runtime['name'],
            is_local=is_local_access(),
            base=request.script_root,
            path=request.path,
            content=commonmark.commonmark(body),
            )

    return http_response(body, headers=headers)


class ActionHandler():
    def _handle_action(self, action, *args, **kwargs):
        handler = getattr(self, action, None) or self.unknown
        return handler(*args, **kwargs)

    def _handle_advanced(func):
        """A decorator function that helps handling an advanced command.

        - Verify POST method.
        - Verify access token.
        - Provide a default return value.
        """
        @functools.wraps(func)
        def wrapper(self,
                localpath,
                query=None,
                format=None,
                *args, **kwargs):
            # require POST method
            if request.method != 'POST':
                headers = {
                    'Allow': 'POST',
                    }
                return http_error(405, 'Method "{}" not allowed.'.format(request.method), format=format, headers=headers)

            # validate and revoke token
            token = query.get('token') or ''

            if not runtime['token_handler'].validate(token):
                return http_error(400, 'Invalid access token.', format=format)

            runtime['token_handler'].delete(token)

            rv = func(
                self=self,
                localpath=localpath,
                query=query,
                format=format,
                *args, **kwargs)

            if rv is not None:
                return rv

            if format:
                return http_response('Command run successfully.', format=format)

            return http_response(status=204)

        return wrapper

    def _handle_lock(func):
        """A decorator function that helps handling the lock.

        - Verify lock name.
        - Verify targetpath.
        """
        @functools.wraps(func)
        def wrapper(self, query=None, format=None, *args, **kwargs):
            # verify name
            name = query.get('name')
            if name is None:
                return http_error(400, "Lock name is not specified.", format=format)

            # validate targetpath
            targetpath = os.path.join(runtime['locks'], name)
            if not targetpath.startswith(os.path.join(runtime['locks'], '')):
                return http_error(400, 'Invalid lock name "{}".'.format(name), format=format)

            return func(
                self=self,
                name=name,
                targetpath=targetpath,
                query=query,
                format=format,
                *args, **kwargs)

        return wrapper

    def _handle_writing(func):
        """A decorator function that helps handling a writing action.
        """
        @functools.wraps(func)
        def wrapper(self, localpath, format=None, *args, **kwargs):
            if os.path.abspath(localpath) == runtime['root']:
                return http_error(403, "Unable to operate the root directory.", format=format)

            return func(
                self=self,
                localpath=localpath,
                format=format,
                *args, **kwargs)

        return wrapper

    def _handle_renaming(func):
        """A decorator function that helps handling a move/copy action.
        """
        @functools.wraps(func)
        def wrapper(self,
                localpath,
                archivefile=None,
                subarchivepath=None,
                query=None,
                format=None,
                *args, **kwargs):
            if archivefile:
                return http_error(400, "File is inside an archive file.", format=format)

            if not os.path.lexists(localpath):
                return http_error(404, "File does not exist.", format=format)

            target = query.get('target')

            if target is None:
                return http_error(400, 'Target is not specified.', format=format)

            targetpath = os.path.normpath(os.path.join(runtime['root'], target.strip('/')))

            if not targetpath.startswith(os.path.join(runtime['root'], '')):
                return http_error(403, "Unable to operate beyond the root directory.", format=format)

            if os.path.lexists(targetpath):
                return http_error(400, 'Found something at target "{}".'.format(target), format=format)

            ta, tsa = get_archive_path(target)
            if ta:
                return http_error(400, "Target is inside an archive file.", format=format)

            return func(
                self=self,
                localpath=localpath,
                targetpath=targetpath,
                format=format,
                *args, **kwargs)

        return wrapper

    def unknown(self, format=None, *args, **kwargs):
        """Default handler for an undefined action"""
        return http_error(400, "Action not supported.", format=format)

    def view(self,
            filepath,
            localpath,
            localtargetpath,
            archivefile=None,
            subarchivepath=None,
            mimetype=None,
            format=None,
            *args, **kwargs):
        """Show the content of a file or list a directory.

        If formatted, show information of the file or directory.
        """
        # show file information for other output formats
        if format:
            if archivefile:
                return handle_subarchive_path(os.path.realpath(archivefile), subarchivepath, mimetype, format=format)

            info = util.file_info(localpath)
            data = {
                'name': info.name,
                'type': info.type,
                'size': info.size,
                'last_modified': info.last_modified,
                'mime': mimetype,
                }
            return http_response(data, format=format)

        # handle directory
        if os.path.isdir(localpath):
            return handle_directory_listing(localtargetpath)

        # handle file
        elif os.path.isfile(localpath):
            # view archive file
            if mimetype in ("application/html+zip", "application/x-maff"):
                return handle_archive_viewing(localtargetpath, mimetype)

            # view markdown
            if mimetype == "text/markdown":
                return handle_markdown_output(filepath, localtargetpath)

            # convert meta refresh to 302 redirect
            if localtargetpath.lower().endswith('.htm'):
                target = util.parse_meta_refresh(localtargetpath).target

                if target is not None:
                    # Keep several chars as javascript encodeURI do,
                    # plus "%" as target may have already been escaped.
                    new_url = urljoin(request.url, quote(target, ";,/?:@&=+$-_.!~*'()#%"))
                    return redirect(new_url)

            # show static file for other cases
            response = static_file(filepath, root=runtime['root'], mimetype=mimetype)

        # handle sub-archive path
        elif archivefile:
            response = handle_subarchive_path(os.path.realpath(archivefile), subarchivepath, mimetype)

        else:
            return http_error(404)

        # don't include charset
        m, p = parse_options_header(response.headers.get('Content-Type'))
        try:
            del p['charset']
        except KeyError:
            pass
        response.headers.set('Content-Type', dump_options_header(m, p))

        return response

    def source(self,
            filepath,
            archivefile=None,
            subarchivepath=None,
            mimetype=None,
            query=None,
            format=None,
            *args, **kwargs):
        """Show file content as plain text."""
        if format:
            return http_error(400, "Action not supported.", format=format)

        if archivefile:
            response = handle_subarchive_path(os.path.realpath(archivefile), subarchivepath, mimetype, list_directory=False)
        else:
            response = static_file(filepath, root=runtime['root'], mimetype=mimetype)

        # show as inline plain text
        # @TODO: Chromium (80) seems to ignore header mimetype for certain types
        #        like image and zip
        encoding = query.get('e', 'utf-8')
        encoding = query.get('encoding', default=encoding)
        response.headers.set('Content-Type', 'text/plain; charset=' + quote(encoding))
        response.headers.set('Content-Disposition', 'inline')

        return response

    def list(self,
            localpath,
            localtargetpath,
            archivefile=None,
            subarchivepath=None,
            query=None,
            format=None,
            *args, **kwargs):
        """List entries in a directory."""
        if not format:
            return http_error(400, "Action not supported.", format=format)

        recursive = query.get('recursive', type=bool)

        if archivefile:
            return handle_zip_directory_listing(os.path.realpath(archivefile), os.path.realpath(archivefile), subarchivepath, recursive=recursive, format=format)

        if os.path.isdir(localpath):
            return handle_directory_listing(localtargetpath, recursive=recursive, format=format)

        return http_error(404, "Directory does not exist.", format=format)

    def static(self,
            filepath,
            format=None,
            *args, **kwargs):
        """Show a static file of the current theme."""
        if format:
            return http_error(400, "Action not supported.", format=format)

        for i in runtime['statics']:
            f = os.path.join(i, filepath)
            if os.path.isfile(f):
                return static_file(filepath, root=i)
        else:
            return http_error(404)

    def edit(self,
            localpath,
            archivefile=None,
            subarchivepath=None,
            query=None,
            format=None,
            *args, **kwargs):
        """Simple text editor for a file."""
        if format:
            return http_error(400, "Action not supported.", format=format)

        if os.path.lexists(localpath) and not os.path.isfile(localpath):
            return http_error(400, "Found a non-file here.", format=format)

        if archivefile:
            with zipfile.ZipFile(archivefile, 'r') as zip:
                try:
                    info = zip.getinfo(subarchivepath)
                except:
                    body = b''
                else:
                    body = zip.read(info)
        else:
            try:
                with open(localpath, 'rb') as f:
                    body = f.read()
            except FileNotFoundError:
                body = b''

        encoding = query.get('e')
        encoding = query.get('encoding', default=encoding)

        try:
            body = body.decode(encoding or 'UTF-8')
        except (LookupError, UnicodeDecodeError):
            encoding = 'ISO-8859-1'
            body = body.decode(encoding)

        body = render_template('edit.html',
                sitename=runtime['name'],
                is_local=is_local_access(),
                base=request.script_root,
                path=request.path,
                body=body,
                encoding=encoding,
                )

        return http_response(body, format=format)

    def editx(self,
            localpath,
            archivefile=None,
            subarchivepath=None,
            mimetype=None,
            format=None,
            *args, **kwargs):
        """HTML editor for a file."""
        if format:
            return http_error(400, "Action not supported.", format=format)

        if os.path.lexists(localpath) and not os.path.isfile(localpath):
            return http_error(400, "Found a non-file here.", format=format)

        if not mimetype in ("text/html", "application/xhtml+xml"):
            return http_error(400, "This is not an HTML file.", format=format)

        if archivefile:
            with zipfile.ZipFile(archivefile, 'r') as zip:
                try:
                    info = zip.getinfo(subarchivepath)
                except:
                    return http_error(404, "File does not exist.", format=format)
        else:
            if not os.path.lexists(localpath):
                return http_error(404, "File does not exist.", format=format)

        body = render_template('editx.html',
                sitename=runtime['name'],
                is_local=is_local_access(),
                base=request.script_root,
                path=request.path,
                )

        return http_response(body, format=format)

    def exec(self, localpath, format=None, *args, **kwargs):
        """Launch a file or directory."""
        if not is_local_access():
            return http_error(400, "Command can only run on local device.", format=format)

        if not os.path.lexists(localpath):
            return http_error(404, "File does not exist.", format=format)

        util.launch(localpath)

        if format:
            return http_response('Command run successfully.', format=format)

        return http_response(status=204)

    def browse(self, localpath, format=None, *args, **kwargs):
        """Open a file or directory in the file browser."""
        if not is_local_access():
            return http_error(400, "Command can only run on local device.", format=format)

        if not os.path.lexists(localpath):
            return http_error(404, "File does not exist.", format=format)

        util.view_in_explorer(localpath)

        if format:
            return http_response('Command run successfully.', format=format)

        return http_response(status=204)

    def config(self, format=None, *args, **kwargs):
        """Show server config."""
        if not format:
            return http_error(400, "Action not supported.", format=format)

        data = runtime['config'].dump_object()

        # filter values for better security
        data = {k:v for k, v in data.items() if k in ('app', 'book')}
        data['app'] = {k:v for k, v in data['app'].items() if k in ('name', 'theme')}

        # add and rewrite values for client to better know the server
        data['app']['base'] = request.script_root
        data['app']['is_local'] = is_local_access()
        data['VERSION'] = __version__;
        data['WSB_DIR'] = WSB_DIR;
        data['WSB_LOCAL_CONFIG'] = WSB_LOCAL_CONFIG;

        return http_response(data, format=format)

    def token(self, format=None, *args, **kwargs):
        """Acquire a token and return its name."""
        return http_response(runtime['token_handler'].acquire(), format=format)

    @_handle_advanced
    @_handle_lock
    def lock(self, name, targetpath, query=None, format=None, *args, **kwargs):
        """Acquire a lock for the given name.

        URL params:
        - chkt: recheck until the lock file not exist or fail out when time out.
        - chks: how long to treat the lock file as stale.
        """
        check_stale = query.get('chks', 300, type=int)
        check_timeout = query.get('chkt', 5, type=int)
        check_expire = time.time() + check_timeout
        check_delta = min(check_timeout, 0.1)

        while True:
            try:
                os.makedirs(targetpath)
            except FileExistsError:
                t = time.time()

                if t >= check_expire or not os.path.isdir(targetpath):
                    return http_error(500, 'Unable to acquire lock "{}".'.format(name), format=format)

                try:
                    lock_expire = os.stat(targetpath).st_mtime + check_stale
                except FileNotFoundError:
                    # Lock removed by another process during the short interval.
                    # Try acquire again.
                    continue

                if t >= lock_expire:
                    # Lock expired. Touch rather than remove and make for atomicity.
                    try:
                        Path(targetpath).touch()
                    except:
                        traceback.print_exc()
                        return http_error(500, 'Unable to regenerate stale lock "{}".'.format(name), format=format)
                    else:
                        break

                time.sleep(check_delta)
            except:
                traceback.print_exc()
                return http_error(500, 'Unable to create lock "{}".'.format(name), format=format)
            else:
                break

    @_handle_advanced
    @_handle_lock
    def unlock(self, name, targetpath, format=None, *args, **kwargs):
        """Release a lock for the given name."""
        try:
            os.rmdir(targetpath)
        except FileNotFoundError:
            pass
        except:
            traceback.print_exc()
            return http_error(500, 'Unable to remove lock "{}".'.format(name), format=format)

    @_handle_advanced
    @_handle_writing
    def mkdir(self,
            localpath,
            archivefile=None,
            subarchivepath=None,
            format=None,
            *args, **kwargs):
        """Create a directory."""
        if os.path.lexists(localpath) and not os.path.isdir(localpath):
            return http_error(400, "Found a non-directory here.", format=format)

        if archivefile:
            try:
                with zipfile.ZipFile(archivefile, 'a') as zip:
                    subarchivepath = subarchivepath + '/'
                    try:
                        info = zip.getinfo(subarchivepath)
                    except KeyError:
                        # subarchivepath does not exist
                        info = zipfile.ZipInfo(subarchivepath, time.localtime())
                        zip.writestr(info, b'', compress_type=zipfile.ZIP_STORED)
            except:
                traceback.print_exc()
                return http_error(500, "Unable to write to this ZIP file.", format=format)

        else:
            try:
                os.makedirs(localpath, exist_ok=True)
            except OSError:
                traceback.print_exc()
                return http_error(500, "Unable to create a directory here.", format=format)

    @_handle_advanced
    @_handle_writing
    def save(self,
            localpath,
            archivefile=None,
            subarchivepath=None,
            query=None,
            format=None,
            *args, **kwargs):
        """Write a file with provided text or uploaded stream."""
        if os.path.lexists(localpath) and not os.path.isfile(localpath):
            return http_error(400, "Found a non-file here.", format=format)

        if archivefile:
            try:
                temp_writing_file = None

                with zipfile.ZipFile(archivefile, 'a') as zip0:
                    # if subarchivepath exists, open a new zip file for writing.
                    try:
                        info = zip0.getinfo(subarchivepath)
                    except KeyError:
                        info = zipfile.ZipInfo(subarchivepath, time.localtime())
                    else:
                        info.date_time = time.localtime()
                        temp_writing_file = archivefile + '.' + str(time_ns())

                    try:
                        with zipfile.ZipFile(temp_writing_file, 'w') if temp_writing_file else zip0 as zip:
                            # write to the zip file
                            file = request.files.get('upload')
                            if file is not None:
                                with zip.open(info, 'w', force_zip64=True) as fh:
                                    stream = file.stream
                                    while True:
                                        s = stream.read(8192)
                                        if not s: break
                                        fh.write(s)
                            else:
                                bytes = query.get('text', '').encode('ISO-8859-1')
                                try:
                                    zip.writestr(info, bytes, compress_type=zipfile.ZIP_DEFLATED, compresslevel=9)
                                except TypeError:
                                    # compresslevel is supported since Python 3.7
                                    zip.writestr(info, bytes, compress_type=zipfile.ZIP_DEFLATED)

                            # copy zip content
                            if temp_writing_file:
                                for info in zip0.infolist():
                                    if info.filename == subarchivepath:
                                        continue

                                    try:
                                        zip.writestr(info, zip0.read(info),
                                                compress_type=info.compress_type,
                                                compresslevel=None if info.compress_type == zipfile.ZIP_STORED else 9)
                                    except TypeError:
                                        # compresslevel is supported since Python 3.7
                                        zip.writestr(info, zip0.read(info),
                                                compress_type=info.compress_type)
                    except:
                        # remove the generated zip file if writing fails
                        if temp_writing_file:
                            try:
                                os.remove(temp_writing_file)
                            except FileNotFoundError:
                                pass

                        raise

                # replace zip0 with the generated zip file
                if temp_writing_file:
                    temp_renaming_file = archivefile + '.' + str(time_ns() + 1)
                    os.rename(archivefile, temp_renaming_file)
                    os.rename(temp_writing_file, archivefile)
                    os.remove(temp_renaming_file)
            except:
                traceback.print_exc()
                return http_error(500, "Unable to write to this ZIP file.", format=format)

        else:
            try:
                os.makedirs(os.path.dirname(localpath), exist_ok=True)
            except:
                traceback.print_exc()
                return http_error(500, "Unable to write to this path.", format=format)

            try:
                file = request.files.get('upload')
                if file is not None:
                    file.save(localpath)
                else:
                    bytes = query.get('text', '').encode('ISO-8859-1')
                    with open(localpath, 'wb') as f:
                        f.write(bytes)
            except:
                traceback.print_exc()
                return http_error(500, "Unable to write to this file.", format=format)

    @_handle_advanced
    @_handle_writing
    def delete(self,
            localpath,
            archivefile=None,
            subarchivepath=None,
            format=None,
            *args, **kwargs):
        """Delete a file or directory."""
        if archivefile:
            try:
                temp_writing_file = archivefile + '.' + str(time_ns())
                deleted = False

                with zipfile.ZipFile(archivefile, 'r') as zip0:
                    try:
                        # copy zip content
                        with zipfile.ZipFile(temp_writing_file, 'w') as zip:
                            for info in zip0.infolist():
                                if (info.filename == subarchivepath or
                                        info.filename.startswith(subarchivepath + '/')):
                                    deleted = True
                                    continue

                                try:
                                    zip.writestr(info, zip0.read(info),
                                            compress_type=info.compress_type,
                                            compresslevel=None if info.compress_type == zipfile.ZIP_STORED else 9)
                                except TypeError:
                                    # compresslevel is supported since Python 3.7
                                    zip.writestr(info, zip0.read(info),
                                            compress_type=info.compress_type)
                    except:
                        # remove the generated zip file if writing fails
                        try:
                            os.remove(temp_writing_file)
                        except FileNotFoundError:
                            pass

                        raise

                # fail since nothing is deleted
                if not deleted:
                    os.remove(temp_writing_file)
                    return http_error(404, "Entry does not exist in this ZIP file.", format=format)

                # replace zip0 with the generated zip file
                temp_renaming_file = archivefile + '.' + str(time_ns() + 1)
                os.rename(archivefile, temp_renaming_file)
                os.rename(temp_writing_file, archivefile)
                os.remove(temp_renaming_file)
            except:
                traceback.print_exc()
                return http_error(500, "Unable to write to this ZIP file.", format=format)

        else:
            if not os.path.lexists(localpath):
                return http_error(404, "File does not exist.", format=format)

            if os.path.islink(localpath):
                try:
                    os.remove(localpath)
                except:
                    traceback.print_exc()
                    return http_error(500, "Unable to delete this link.", format=format)
            elif os.path.isfile(localpath):
                try:
                    os.remove(localpath)
                except:
                    traceback.print_exc()
                    return http_error(500, "Unable to delete this file.", format=format)
            elif os.path.isdir(localpath):
                try:
                    try:
                        # try rmdir for a possible windows directory junction,
                        # which is not detected by os.path.islink
                        os.rmdir(localpath)
                    except OSError:
                        # directory not empty
                        shutil.rmtree(localpath)
                except:
                    traceback.print_exc()
                    return http_error(500, "Unable to delete this directory.", format=format)

    @_handle_advanced
    @_handle_writing
    @_handle_renaming
    def move(self,
            localpath,
            targetpath,
            format=None,
            *args, **kwargs):
        """Move a file or directory."""
        os.makedirs(os.path.dirname(targetpath), exist_ok=True)

        try:
            os.rename(localpath, targetpath)
        except:
            traceback.print_exc()
            return http_error(500, 'Unable to move to the target.', format=format)

    @_handle_advanced
    @_handle_writing
    @_handle_renaming
    def copy(self,
            localpath,
            targetpath,
            format=None,
            *args, **kwargs):
        """Copy a file or directory."""
        os.makedirs(os.path.dirname(targetpath), exist_ok=True)

        try:
            try:
                shutil.copytree(localpath, targetpath)
            except NotADirectoryError:
                shutil.copy2(localpath, targetpath)
        except:
            traceback.print_exc()
            return http_error(500, 'Unable to copy to the target.', format=format)

    _handle_advanced = staticmethod(_handle_advanced)
    _handle_lock = staticmethod(_handle_lock)
    _handle_writing = staticmethod(_handle_writing)
    _handle_renaming = staticmethod(_handle_renaming)

action_handler = ActionHandler()


@bp.route('/', methods=['GET', 'HEAD', 'POST'])
@bp.route('/<path:filepath>', methods=['GET', 'HEAD', 'POST'])
def handle_request(filepath=''):
    """Handle an HTTP request (HEAD, GET, POST).
    """
    # replace SCRIPT_NAME with the custom if set
    if runtime['config']['app']['base']:
        request.environ['SCRIPT_NAME'] = runtime['config']['app']['base']

    query = request.values

    action = query.get('a', default='view')
    action = query.get('action', default=action)

    format = query.get('f')
    format = query.get('format', default=format)

    # handle authorization
    perm = get_permission(request.authorization)
    if not verify_authorization(perm, action):
        response = http_error(401, "You are not authorized.", format=format)
        response.www_authenticate.set_basic("Authentication required.")
        return response

    # determine primary variables
    #
    # filepath: the URL path below app base (not percent encoded)
    # localpath: the file system path corresponding to filepath
    # localtargetpath: localpath with symbolic link resolved
    # mimetype: the mimetype from localtargetpath
    # archivefile: the file system path of the ZIP archive file, or None
    # subarchivepath: the URL path below archivefile (not percent encoded)
    localpath = os.path.abspath(os.path.join(runtime['root'], filepath.strip('/')))
    localtargetpath = os.path.realpath(localpath)
    archivefile, subarchivepath = get_archive_path(filepath)
    mimetype, _ = mimetypes.guess_type(localtargetpath)

    return action_handler._handle_action(
        action=action,
        filepath=filepath,
        localpath=localpath,
        localtargetpath=localtargetpath,
        archivefile=archivefile,
        subarchivepath=subarchivepath,
        mimetype=mimetype,
        query=query,
        format=format,
        )


def make_app(root=".", config=None):
    if not config:
        config = Config()
        config.load(root)

    # runtime variables
    _runtime = {}
    _runtime['config'] = config
    _runtime['root'] = os.path.abspath(os.path.join(root, config['app']['root']))
    _runtime['name'] = config['app']['name']

    # add path for themes
    _runtime['themes'] = [
        os.path.join(_runtime['root'], WSB_DIR, 'themes', config['app']['theme']),
        os.path.join(os.path.dirname(__file__), 'themes', config['app']['theme']),
        ]
    _runtime['statics'] = [os.path.join(t, 'static') for t in _runtime['themes']]
    _runtime['templates'] = [os.path.join(t, 'templates') for t in _runtime['themes']]

    _runtime['tokens'] = os.path.join(_runtime['root'], WSB_DIR, 'server', 'tokens')
    _runtime['locks'] = os.path.join(_runtime['root'], WSB_DIR, 'server', 'locks')

    # init token_handler
    _runtime['token_handler'] = util.TokenHandler(_runtime['tokens'])

    # main app instance
    app = flask.Flask(__name__, instance_path=_runtime['root'])
    app.register_blueprint(bp)
    app.config['WEBSCRAPBOOK_RUNTIME'] = _runtime

    xheaders = {
            'x_for': config['app']['allowed_x_for'],
            'x_proto': config['app']['allowed_x_proto'],
            'x_host': config['app']['allowed_x_host'],
            'x_port': config['app']['allowed_x_port'],
            'x_prefix': config['app']['allowed_x_prefix'],
            }

    if any(v for v in xheaders.values()):
        app.wsgi_app = ProxyFix(app.wsgi_app, **xheaders)

    app.jinja_loader = jinja2.FileSystemLoader(_runtime['templates'])
    app.jinja_env.globals.update({
            'os': os,
            'time': time,
            'get_breadcrumbs': util.get_breadcrumbs,
            'format_filesize': util.format_filesize,
            'quote_path': quote_path,
            })

    return app
